# -*- coding: utf-8 -*-
"""
This module contains functions to generate config file to be used within BMSS2.

Features:
-to generate SBML string from a online database
-to generate Config File from SBML string
-to generate Setting Template File from SBML string for BMSS simulation
"""

import tellurium as te, os
import numpy as np
import tempfile
import re
import io
import zipfile
import requests
from bs4 import BeautifulSoup 
from pathlib import Path
from synbiopython import genbabel
import synbiopython as syn
#from synbiopython.genbabel import utilities

_tspan = str(list(np.linspace(0, 1000, 51)))

#--- Get Online Model From Website ---
def get_online_biomodel(Biomodels_ID, **kwargs):
    modelstr = syn.genbabel.SEDMLOMEXgen.get_sbml_biomodel(Biomodels_ID, **kwargs)
    return modelstr

#--- Main Bodys ---
def sbmltoconfig(sbmlstr, system_type, output_path, tspan=None):
    '''
    Generates the config file statment and nested dictionary.
    
    :param sbmlstr: SBML string in string format
    :param system_type: System Type of Model defined in String Format
    :param tspan: Tspan of Model defined in String Format
    :param Model_name: Name of Model defined in String Format, Used to name files
    :return: config file in string format and nested dictionary for settings
    
    '''
    Model_name = system_type
    tspan      = _tspan if tspan is None else str(list(tspan))
    
    #--Ensure Tspan declared correctly--
    tspanchecker(tspan)
    #--Ensure that sbmlstr can be converted--
    configpossiblechecker(sbmlstr)
    
    species_store, parameters_store, reactions_store, rules_store, antimony_str, description_store, reference_store = gen_dictionary(sbmlstr)

    
    species_clean           = clean_id(species_store)
    parameter_clean         = clean_id(parameters_store)
    speciesvalue_clean      = gen_cleanvalue(species_store)
    parametersvalue_clean   = gen_cleanvalue(parameters_store)
    units_clean             = clean_units(parameters_store)
    
    species_dict            = build_settings(species_clean, speciesvalue_clean)
    parameter_dict          = build_settings(parameter_clean, parametersvalue_clean)
    parametersunits_dict    = build_settings(parameter_clean, units_clean)
    equations               = gen_eqns(antimony_str)
    try:
        species_parameter_descriptions = gen_species_parameter_descriptions(antimony_str, species_dict, parameter_dict)
        print('Species and Parameter Descriptions exist\n')
    except ValueError:
        species_parameter_descriptions = ''
       
    settings_combine = {
        "Species" : species_dict,
        "parameters" : parameter_dict,
        "units" : parametersunits_dict,
        "equations" : equations,
        "description": description_store,
        "reference": reference_store,
        "species_parameter_descriptions": species_parameter_descriptions
        }
    
    config_statement    = gen_config(settings_combine, system_type, tspan)
    settings_statement  = gen_settingstemplate(settings_combine, system_type, tspan, Model_name)
    filepath            = '_'.join( [i.strip() for i in Model_name.split(',')] )
    config_filepath     = filepath + '.ini'
    config_filepath     = Path(output_path) / config_filepath  
    settings_filepath   = filepath + '_sim_settings.ini'
    settings_filepath   = Path(output_path) / settings_filepath
    # Config_filepath     = os.path.join(str(output_path), str(Model_name) + '_coremodel.ini')
    # settings_filepath   = os.path.join(str(output_path), 'sim_settings_' + str(Model_name) + '.ini')
    
    with open(config_filepath, "w") as f:
        f.write(config_statement) 
    with open(settings_filepath, "w") as f:
        f.write(settings_statement) 

    print()
    print('The config file path:')
    print(config_filepath)
    print()
    print('The settings template file path:')
    print(settings_filepath)
    
    return config_statement, settings_combine

def gen_config(settings, system_type, tspan):
    '''
    Generates config statment for config file
    
    :param settings: Nested Dictionary for all settings
    :param system_type: System Type of Model defined in String Format
    :param tspan: Tspan of Model defined in String Format
    :return: full config statement in string format
    
    '''
    systemtype_section      = '''[system_type]
system_type = ''' + system_type + '\n'
    
    statestop_section       = '''
[states]
states = ''' 
    statestop_section       = statestop_section + build_section_top(settings, 'Species')
    
    parametertop_section    = '''
[parameters]
parameters = '''
    parametertop_section = parametertop_section + build_section_top(settings, 'parameters')
    
    inputs_section          = '''
[inputs]
inputs = 
'''
    
    equations_section       = '''
[equations]
equations =
''' 
    equations_section       = build_section(equations_section, settings, 'equations')

    
    combined_top            = (systemtype_section + statestop_section + parametertop_section + inputs_section 
                            + equations_section + settings['species_parameter_descriptions'] +
                            settings['description'] + settings['reference'])
    
    species_section         = 'init = \n'
    species_section         = build_section(species_section, settings, 'Species')
    parameters_section      = 'parameter_values = \n'
    parameters_section      = build_section(parameters_section, settings, 'parameters')
    units_section           = 'units = \n'
    
    #--Checker to ensure all parameters have units defined)
    if len(settings['units']) != len(settings['parameters']):
        raise Exception('Missing Unit for Parameter')
        
    units_section               = build_section(units_section, settings, 'units')
    priors, parameter_bounds    = gen_prior_bounds(settings)
    
    
    combined_bottom             = ('[_]\n' + 'system_type = '+ system_type + ' \n\n' + species_section + parameters_section 
                                   + priors + parameter_bounds + units_section)
    combined                    = combined_top + combined_bottom + 'tspan = \n' + ' '*4 + tspan
    
    return combined

#--- Settings Template Creator ---

def gen_settingstemplate(settings, system_type, tspan, Model_name):
    '''
    Generates settings template for Tutorial 2 Simulation
    
    :param settings: Nested Dictionary for all settings
    :param system_type: System Type of Model defined in String Format
    :param tspan: Tspan of Model defined in String Format
    :param Model_name: Name of Model defined in String Format, Used to name files
    :return: full settings template in string format
    
    '''
    settings_template           = '[' + system_type + ']\n\n'
    species_section             = 'init = \n'
    species_section             = build_section(species_section, settings, 'Species')
    parameters_section          = 'parameter_values = \n'
    parameters_section          = build_section(parameters_section, settings, 'parameters')
    units_section               = 'units = \n'
    units_section               = build_section(units_section, settings, 'units')
    priors, parameter_bounds    = gen_prior_bounds(settings)
    settings_template           = (settings_template + species_section + parameters_section
                                   + priors + parameter_bounds + units_section
                                   + 'tspan = \n' + ' '*4 + tspan)
    
    
    
    return settings_template

# --- XML Parse to readable data ---
def gen_dictionary(data):
    '''
    Passes XML data through beautifulsoup parser
    making data more easily readable
    
    :param data: SBML string in XML string format
    :returns: all required data in list format

    '''

    antimony_str            = te.sbmlToAntimony(data)
 
    Bs_data                 = BeautifulSoup(data, "xml") 
    b_species               = Bs_data.find_all('listOfSpecies')
    species_store           = Convert(str(b_species))
    b_parameters            = Bs_data.find_all('listOfParameters')
    parameters_store        = Convert(str(b_parameters))
    b_reactions             = Bs_data.find_all('listOfReactions')
    reactions_store         = Convert(str(b_reactions))
    b_rules                 = Bs_data.find_all('listOfRules')
    rules_store             = Convert(str(b_rules)) 
    #soup.find_all("a", class_="sister")
    #print(Bs_data.find_all('div', class_='dc:description'))
    if Bs_data.find_all(class_='dc:description') != []:
        b_description       = Bs_data.find_all(class_='dc:description')
        description_store   = Convert(str(b_description))
        description_store   = gen_description(description_store)
    else:
        description_store   = ''
    if Bs_data.find_all(class_='dc:bibliographicCitation') != []:
        reference_store     = gen_reference(Bs_data)
    else:
        reference_store     = ''        
        
    '''
    print(species_store, '\n')
    print(parameters_store, '\n')
    print(reactions_store, '\n')
    print(rules_store, '\n')
    '''
    #print(description_store)
    return species_store, parameters_store, reactions_store, rules_store, antimony_str, description_store, reference_store

#--- ID,Values Clean-up---
#Finds the ID of Species/Paramters and combines with values

def clean_id(store):
    '''
    Finds the IDs within the parsed data
    
    :param store: stored Species/Parameters data in list format
    :return: IDs for Species/Parameters in list format
    
    '''
    tempstore   = ''
    store_clean = []
    for value in store:
        if 'listOfSpecies' not in value and 'listOfParameters' not in value:
            #if 'name=' in value:
            if 'id=' in value:
                id_find     = value.index('id="')
                tempstore   = value[id_find+4:] 
                id_end      = tempstore.index('"')
                store_clean.append(tempstore[:id_end])
    #print(store_clean)
    return store_clean

def clean_value(value, store_clean):
    '''
    Finds the values within parsed data
    
    :param value: Line from parsed data in string format
    :param store_clean: Existing Values for Species/Parameters in list format
    :return: New added values for Species/Parameters in list format
    
    '''
    tempstore = ''
    if 'initialAmount=' in value:
        id_find     = value.index('initialAmount="')
        tempstore   = value[id_find+15:] 
        #print('this is tempstore\n', tempstore)
        id_end      = tempstore.index('"')
        store_clean.append(tempstore[:id_end])    
    elif 'initialConcentration=' in value:
        id_find     = value.index('initialConcentration="')
        tempstore   = value[id_find+22:] 
        id_end      = tempstore.index('"')
        store_clean.append(tempstore[:id_end])                  
    elif 'value=' in value:
        id_find     = value.index('value="')
        tempstore   = value[id_find+7:] 
        id_end      = tempstore.index('"')
        store_clean.append(tempstore[:id_end])

    return store_clean

def gen_cleanvalue(store):
    '''
    Fills in '0' amount for Species and Parameters
    without initial values
    
    :param store: List containing parsed data from Species/Parameters
    :return: Full Values for Species/Parameters in list format
    
    '''
    store_clean = []
    for value in store:
        #print(value)
        #if 'listOfSpecies' not in value and 'listOfParameters' not in value:
        if 'listOfParameters' in store[0] and 'id="' in value:
            len_store_clean     = len(store_clean)
            store_clean         = clean_value(value, store_clean)
            if len_store_clean == len(store_clean):
                store_clean.append('0')
        else:
            store_clean = clean_value(value, store_clean)
                
    #print('this is store_clean\n', store_clean)
    return store_clean

def clean_units(store):
    '''
    Finds the Units for Parameters if declared
    Feels in default units if not delcared (per_sec)
    
    :param store: List containing parsed data from Species/Parameters
    :return: full units in list format
    
    '''
    tempstore = ''
    store_clean = []
    for value in store:
        if 'listOfSpecies' not in value and 'listOfParameters' not in value:
            if 'substanceUnits' in value:
                id_find     = value.index('substanceUnits="')
                tempstore   = value[id_find+16:] 
                id_end      = tempstore.index('"')
                store_clean.append(tempstore[:id_end])
            elif 'units' in value:
                id_find     = value.index('units="')
                tempstore   = value[id_find+7:]
                tempstore   = tempstore.replace('_le', '_less')
                id_end      = tempstore.index('"')
                store_clean.append(tempstore[:id_end])
            else:
                store_clean.append('per_sec')
                
    return store_clean  
  
def build_settings(idstore, valuestore):
    '''
    Combines the IDs with their respective values
    
    :param idstore: all IDs in list format
    :param valuestore: all values of IDS in list format
    :return: matched ID to values in dictionary format
    
    '''
    settings_dict   = {}
    i               = 0
    for value in idstore:
        settings_dict[value] = valuestore[i]
        i += 1
    return settings_dict


#--- Equations/Reactions ---
def gen_eqns(antimony_str):
    '''
    Generates all equations from assignment rules, rate rules and reactions
    
    :param antimony_str: antimony string in string format
    :returns: all equations in list format
    
    '''
    eqn_clean = ['']
    if '// Assignment Rules:' in antimony_str:
        assignment_find         = antimony_str.index('// Assignment Rules:')
        assignment_end          = antimony_str.find('\n\n', assignment_find)
        tempstore_assignment    = antimony_str[assignment_find+21:assignment_end]
        tempstore_assignment    = eqnreplace(tempstore_assignment)
        assignment_clean        = Convert(tempstore_assignment)
        
        eqn_clean = assignment_clean +['']
        
    if '// Rate Rules:' in antimony_str:
        
        rate_find       = antimony_str.index('// Rate Rules:')
        rate_end        = antimony_str.find('\n\n', rate_find)
        tempstore_rate  = antimony_str[rate_find+15:rate_end]
        tempstore_rate  = eqnreplace(tempstore_rate)
        tempstore_rate  = tempstore_rate.replace("  ", '  d')
        rate_clean      = Convert(tempstore_rate)
        
        eqn_clean       = eqn_clean + [''] + rate_clean
        
    if '// Reactions:' in antimony_str:
        reaction_formula_store  = gen_reactions(antimony_str)
        before                  = len(reaction_formula_store)
        after                   = 0
        reaction_formula_store = sorted(reaction_formula_store)
        
        while before != after:
            before                  = len(reaction_formula_store)
            reaction_formula_store  = clean_reactions(reaction_formula_store)
            after                   = len(reaction_formula_store)
        
        #eqn_clean = eqn_clean + ['']
        for eqn in reaction_formula_store:
            eqn                     = eqnreplace(eqn)
            eqn_clean.append(eqn)

    
    return eqn_clean

def eqnreplace(tempstore):
    '''
    Removes unnecessary keys in reaction statement
    
    :param tempstore: equation statement in strign format
    :return: clean statement in string format
    
    '''
    tempstore = tempstore.replace(';', '')
    tempstore = tempstore.replace('^', '**')
    tempstore = tempstore.replace(':', '')
    tempstore = tempstore.replace("'", '')
    tempstore = tempstore.replace(")(", ')*(')
    return tempstore

def clean_eqns(store_species):
    '''
    Converts Reaction statements in antimony str into format in config file. Returns a multiplier for reaction
    
    :param store_species: ID of species in reaction
    :returns store_species: list format if reactions have more than 1 variable in reaction
    :returns store_species: string format if reactions only have 1 variable in reaction
    :returns species_multiplier: list format if reactions have more than 1 variable in reaction
    :returns species_multiplier: string format if reactions only have 1 variable in reaction
    
    '''
    if store_species.count('+') > 0:
        store_species       = Convertplus(store_species)
        species_multiplier  = []
        multiplier_remove   = ''
        for index, item in enumerate(store_species):
            store_species[index]    = item.replace(" + ", '')
            store_species[index]    = store_species[index].replace("$", '')
            if bool(re.search(r'\ [\d\.\d]+ ', store_species[index])) == True:
                species_multiplier.append(re.search(r'\ [\d\.\d]+ ', store_species[index]).group())
                multiplier_remove   = str(re.search(r'\ [\d\.\d]+ ', store_species[index]).group())
                species_multiplier[index] = species_multiplier[index].replace(' ', '')
                multiplier_remove   = multiplier_remove.replace(' ', '')
            else:
                species_multiplier.append('')
            store_species[index]    = store_species[index].replace(" ", '')
            store_species[index]    = 'd' + store_species[index]
            store_species[index]    = store_species[index].replace('$', '')
            store_species[index]    = store_species[index].replace(str(multiplier_remove), '')

    else:
        species_multiplier      = ''
        multiplier_remove       = ''
        store_species           = store_species.replace("+", '')
        store_species           = store_species.replace('$', '')
        if bool(re.search(r'\ [\d\.\d]+ ', store_species)) == True:
            species_multiplier  = re.search(r'\ [\d\.\d]+ ', store_species).group()
            multiplier_remove   = str(re.search(r'\ [\d\.\d]+ ', store_species).group())
            species_multiplier  = species_multiplier.replace(' ', '')
            multiplier_remove   = multiplier_remove.replace(' ', '')
            
        store_species           = store_species.replace(' ', '')
        store_species           = 'd' + store_species
        store_species           = store_species.replace(str(multiplier_remove), '')

    
    return store_species, species_multiplier

def get_posreaction(reaction_formula_store, value, reaction_reactant_end, reaction_product_end, reactioneqns):
    '''
    Gets the reaction equation is the reactions creates only a product
    
    :param reaction_formula_store: List containing all reactions
    :param value: Line of equation in reaction in string format
    :param reaction_reactant_end: Index number of the position of the start of the ID
    :param reaction_product_end: Index number of the position of the end of the ID
    :param reactioneqns: reaction equation in string format
    :returns: added "create product" reactions in list format
    
    '''
    tempstore_product                           = value[reaction_reactant_end+2:reaction_product_end]
    tempstore_product_list, species_multiplier  = clean_eqns(tempstore_product)
    if isinstance(tempstore_product_list, list) == True:
        for index, item in enumerate(tempstore_product_list):
            if species_multiplier[index] != '':
                reaction_formula_store.append('  ' + item + ' = +(1/' + species_multiplier[index] 
                                              + ')' + '(' + reactioneqns + ')')
            else:
                reaction_formula_store.append('  ' + item + ' = +(' + reactioneqns + ')')
    else:
            if species_multiplier != '':
                reaction_formula_store.append('  ' + tempstore_product_list + ' = +(1/' 
                                              + species_multiplier + ')' + '(' + reactioneqns + ')')
            else:
                reaction_formula_store.append('  ' + tempstore_product_list 
                                              + ' = +(' + reactioneqns + ')')
    
    return reaction_formula_store 

def get_negreaction(reaction_formula_store, value, reaction_reactant_start, reaction_reactant_end, reactioneqns):
    '''
    Gets the reaction equation is the reactions shows a usage of a reactant
    
    :param reaction_formula_store: List containing all reactions
    :param value: Line of equation in reaction in string format
    :param reaction_reactant_start: Index number of the position of the start of the ID
    :param reaction_reactant_end: Index number of the position of the end of the ID
    :param reactioneqns: reaction equation in string format
    :returns: added "use reactant" reactions in list format
    
    '''
    tempstore_reactant                              = value[reaction_reactant_start+1:reaction_reactant_end]
    tempstore_reactant_list, species_multiplier     = clean_eqns(tempstore_reactant)
    if isinstance(tempstore_reactant_list, list) == True:
        for index, item in enumerate(tempstore_reactant_list):
            if species_multiplier[index] != '':
                reaction_formula_store.append('  ' + item + ' = -(1/' + species_multiplier[index] 
                                              + ')' +'(' + reactioneqns + ')')
            else:
                reaction_formula_store.append('  ' + item + ' = -(' + reactioneqns + ')')
    else:
        if species_multiplier != '':
            reaction_formula_store.append('  ' + tempstore_reactant_list + ' = -(1/' + species_multiplier 
                                          + ')' + '(' + reactioneqns + ')')
        else:
            reaction_formula_store.append('  ' + tempstore_reactant_list + ' = -(' + reactioneqns + ')')
    return reaction_formula_store
 
def gen_reactions(antimony_str):
    '''
    Generates the all rate rules equations from reactions from antimony string
    
    :param antimony_str: antimony string in string format
    :returns: all reactions in list format
    
    '''
    reactioneqns                    = ''
    reaction_formula_store          = []
    reaction_find                   = antimony_str.index('// Reactions:')
    reaction_start                  = antimony_str.find('\n', reaction_find)
    reaction_end                    = antimony_str.find('\n\n', reaction_find)
    tempstore_reaction              = antimony_str[reaction_start+1:reaction_end]
    tempstore_reaction              = Convert(tempstore_reaction)
    for value in tempstore_reaction:
        reaction_reactant_start     = value.find(':')
        reaction_reactant_end       = value.find('=>')
        if reaction_reactant_end    == -1:
            reaction_reactant_end   = value.find('->')
        reaction_product_end        = value.find(';')
        reaction_formula_end        = value.find(';', reaction_product_end+1)
        reactioneqns                = value[reaction_product_end+2:reaction_formula_end]
        #Replaces compartment name in reactions
        if '// Compartments' in antimony_str:
            compartment_name        = clean_compartment(antimony_str)
            reactioneqns            = reactioneqns.replace(compartment_name, '1')
  
        if value[reaction_reactant_start+1:reaction_reactant_end] != "  " and value[reaction_reactant_end+2:reaction_product_end] == " ":
            #print('only reactant present in reaction statement\n')
            reaction_formula_store  = get_negreaction(reaction_formula_store, value, reaction_reactant_start, reaction_reactant_end, reactioneqns)

        if value[reaction_reactant_start+1:reaction_reactant_end] == "  " and value[reaction_reactant_end+2:reaction_product_end] != " ":
            #print('only product present in reaction statement\n')
            reaction_formula_store  = get_posreaction(reaction_formula_store, value, reaction_reactant_end, reaction_product_end, reactioneqns)
           
        if value[reaction_reactant_start+1:reaction_reactant_end] != "  " and value[reaction_reactant_end+2:reaction_product_end] != " ":
            #print('reactant and product present in reaction statement\n')
            reaction_formula_store  = get_negreaction(reaction_formula_store, value, reaction_reactant_start, reaction_reactant_end, reactioneqns)
            reaction_formula_store  = get_posreaction(reaction_formula_store, value, reaction_reactant_end, reaction_product_end, reactioneqns)
    
    return reaction_formula_store

def clean_reactions(reaction):
    '''
    Cleans up reactions by combining equations pertaining to the same species
    
    :param reaction: Raw equations before cleaning in list format
    :return: cleaned reactions in list format
    
    '''
    
    new_equations       = sorted(reaction)
    equations_final     = []
    for index, item in enumerate(new_equations):
        equation_before_start       = new_equations[index-1].find('d')
        equation_before_end         = new_equations[index-1].find('=')
        equation_before             = new_equations[index-1][equation_before_start:equation_before_end]
        equation_current_start      = item.find('d')
        equation_current_end        = item.find('=')
        equation_current            = item[equation_current_start:equation_current_end]
        if (index+1) != len(new_equations):
            equation_after_start    = new_equations[index+1].find('d')
            equation_after_end      = new_equations[index+1].find('=')        
            equation_after          = new_equations[index+1][equation_after_start:equation_after_end]
        else:
            equation_after          = 'null'
            
        if equation_current == equation_before:
            new_equation            = str(new_equations[index-1]) + str(item[equation_current_end+1:])
            equations_final.append(new_equation)
        elif equation_current != equation_before and equation_current != equation_after:
            equations_final.append(item)

    return equations_final

def clean_compartment(antimony_str):
    '''
    Finds the compartment to remove it from equations for simplicity
    
    :param antimony_str: antimony string in string format
    :returns: compartment name in string format
    
    '''
    compartment_find    = antimony_str.index('// Compartments')
    compartment_start   = antimony_str.find('t ', compartment_find)
    compartment_end     = antimony_str.find(';', compartment_start)
    compartment_name    = antimony_str[compartment_start+2:compartment_end]
    #print(compartment_name)
    return compartment_name


# --- Section Creators ---
def build_section(section, settings, key):
    '''
    Builds Section according to key
    
    :param section: config file statement in string format
    :param settings: settings in nested dictionary format
    :param key: variable that determines which if statement to follow, string format
    :return: updated section in string format
    
    '''
    indent = ' '*4
    if key == 'units':
        for species in settings[key].keys():
            section     = section + indent + str(species) + ' = '  + str(settings[key].get(species))
            if species == list(settings[key])[-1]:
                section = section + '\n\n'
            else:
                section = section + ',\n'
    elif key == 'equations':
        for species in settings[key]:
            section = section + (' '*2) + str(species)
            if species == list(settings[key])[-1]:
                section = section + '\n\n'
            else:
                section = section + '\n'  
    else:
        for species in settings[key].keys():
            section = section + indent + str(species) + ' = ['  + str(settings[key].get(species)) +']'
            if species == list(settings[key])[-1]:
                section = section + '\n\n'
            else:
                section = section + ',\n'
    
    return section

def build_section_top(settings, key):
    '''
    Build section for top only of config file
    
    :param settings: settings in nested dictionary format
    :param key: variable that determines which if statement to follow, string format
    :return: top section in string format
    
    '''
    statetemp = ''
    for species in settings[key].keys():
        if species == list(settings[key])[-1]:
            statetemp = statetemp + species + '\n'
        else:
            statetemp = statetemp + species + ', '
    
    return statetemp

def gen_prior_bounds(settings):
    '''
    Generates Priors and Bounds from settings
    
    :param settings: settings in nested dictionary format
    :return: priors and bounds in string format
    
    '''
    priors = ''
    bounds = 'parameter_bounds = \n'
    
    for parameter in settings['parameters']:
        upperbound_value    = float(settings['parameters'][parameter]) + 0.5
        bounds              = bounds + ' '*4 + str(parameter) + ' = [0, ' 
        if parameter != list(settings['parameters'])[-1]:
            if upperbound_value > 0.5:
                upperbound  = str(round((upperbound_value)))
                bounds      = bounds + upperbound + '],\n'
            else:
                bounds      = bounds + '1],\n'    
            #print(settings['parameters'][parameter])
        else:
            if upperbound_value > 0.5:
                upperbound  = str(round((upperbound_value)))
                bounds      = bounds + upperbound + ']\n\n'
            else:
                bounds      = bounds + '1]\n\n'
                
    
    ''' #Example statement
priors = 
	degm  = [0.015, 0.05],
	degp  = [0.012, 0.04]

parameter_bounds = 
	k_ind = [1e-3, 1],
	synm  = [1e-6, 1e-4],
	degm  = [0.01, 0.5],
	synp  = [1e-3, 1],
	degp  = [1e-3, 0.3],
	ind   = [0, 1]   
    ''' 
    
    return priors, bounds

def gen_description(desc_list):
    '''
    Generates Description section
    
    :param desc_list: parsed description data from XML in list format 
    :return: cleaned description in string format
    
    '''
    clean_string = '[descriptions]\n' + 'Description = '
    clean_string = clean_string + xmlclean(desc_list)
    clean_string = clean_string + '\n\n'
    return clean_string 

def gen_reference(Bs_data):
    '''
    Generates Reference section
    
    :param string: parsed description data from XML 
    :return: cleaned reference in string format
    
    '''
    clean_string    = 'Reference= \n'
    title           = Bs_data.find_all(class_='bibo:title')
    title           = str(xmlclean(str(title)))
    title_start     = title.find('">')
    title_end       = title.find('.', title_start)
    clean_title     = title[title_start+2:title_end+1]
    authors         = Bs_data.find_all(class_='bibo:authorList')
    authors         = xmlclean(str(authors))
    journal         = Bs_data.find_all(class_='bibo:Journal')
    journal         = xmlclean(str(journal))
    doi             = Bs_data.find_all(class_='bibo:doi')
    doi             = xmlclean(str(doi))
    clean_string    = clean_string + ' '*4 + 'title: ' + clean_title +'\n'
    clean_string    = clean_string + ' '*4 + 'authors: ' + authors +'\n'
    clean_string    = clean_string + ' '*4 + 'journal: ' + journal +'\n'
    clean_string    = clean_string + ' '*4 + 'doi: ' + doi +'\n'
    clean_string    = clean_string + '\n\n'
    return clean_string 

def xmlclean(xmldata):
    '''
    Cleans away xml related characters
    
    :param xmldata: parsed data from XML 
    :return: cleaned text in string format
    
    '''
    clean_string = ''
    if isinstance(xmldata, list) == True:
        for value in xmldata:
            value = value.replace('<p>', '')
            value = value.replace('</a>', '')
            value = value.replace('</p>', '')
            value = value.replace('</div>', '')
            value = value.replace('.', '. ')
            value = value.replace('  ', ' ')
            value = value.replace('[', '')
            value = value.replace(']', '')
            value = value.replace('<div class="dc:description">', '')
            value = value.replace('<div class="bibo:title">', '')
            value = value.replace('<div class="dc:bibliographicCitation">', '')
            value = value.replace('<div class="bibo:authorList">', '')
            value = value.replace('<div class="bibo:Journal">', '')
            value = value.replace('<div class="bibo:authorList">', '')
            clean_string = clean_string + value
    else:
        value = xmldata
        value = value.replace('<p>', '')
        value = value.replace('\n', ' ')
        value = value.replace('  ', '')
        value = value.replace('</a>', '')
        value = value.replace('</p>', '')
        value = value.replace('</div>', '')
        value = value.replace('[', '')
        value = value.replace(']', '')
        value = value.replace('<div class="dc:description">', '')
        value = value.replace('<div class="bibo:title">', '')
        value = value.replace('<div class="dc:bibliographicCitation">', '')
        value = value.replace('<div class="bibo:authorList">', '')
        value = value.replace('<div class="bibo:Journal">', '')
        value = value.replace('<div class="bibo:authorList">', '')
        clean_string = value
    return clean_string

def gen_species_parameter_descriptions(antimony_str, species_dict, parameter_dict):
    '''
    Generates Species and Parameter Descriptions if defined.
    
    :param antimony_str: SBML Model converted to Antimony format string 
    :param species_dict: dictionary containing all species in model
    :param parameter_dict: dictionary containing all species in model
    :return: species and parameter description in string format
    
    '''
    name_start      = antimony_str.index('// Display Names:')
    name_end        = antimony_str.find('\n\n', name_start)
    display_names   = antimony_str[name_start:name_end]
    display_names   = Convert(display_names)
    s_description   = 'Definition of states=\n'
    p_description   = 'Definition of parameters=\n'
    for line in display_names:
        try:
            variablename    = line.index('is')
        except ValueError:
            variablename = 2
        variable            = line[:variablename]
        variable            = variable.replace('  ', '')
        variable            = variable.replace(' ', '')
        #print(variable)
        if variable in species_dict.keys():
            line            = line.replace(' is "', ': ')
            line            = line.replace('";', '')
            s_description   = s_description + ' '*2 + line + '\n'
        if variable in parameter_dict.keys():
            line            = line.replace(' is "', ': ')
            line            = line.replace('";', '')
            p_description   = p_description + ' '*2 + line + '\n'
    s_description           = s_description + '\n'
    p_description           = p_description + '\n'
    #print(s_description)
    #print(p_description)
    s_p_descriptions        = s_description + p_description
    
    return s_p_descriptions
    
    
#--- Other Functions ---
def Convert(string): 
    '''
    Converts string into list
    New row based on occurence of "\\n"
    
    '''
    li = list(string.split("\n")) 
    return li 

def Convertplus(string): 
    '''
    Converts string into list
    New row based on occurence of "+"
    
    '''
    li = list(string.split("+"))
    return li

def tspanchecker(tspan):
    '''
    :meta private:
    '''
    if 'linspace(' == tspan[:8] and ')' == tspan[-1]:
        args = tspan[8:len(tspan)-1]
        try:
            args            = [int(i) for i in args.split(',')]
            start, stop, *n = args
            if start >= stop:
                raise Exception('Invalid tspan. Start must be less than stop.')
        except Exception as e:
            raise e
            
    if tspan[0] != '[':
        raise Exception('Include "[" at the start of tspan')
    elif tspan[-1] != ']':
        raise Exception('Include "]" at the end of tspan')
        
    vals = tspan[1: len(tspan)-1]
    vals = [float(i) for i in vals.split(',')]
    vals = np.unique(vals)
    return str(list(vals))

def configpossiblechecker(sbmlstr):
    '''
    :meta private:
    '''
    antimony_str    = te.sbmlToAntimony(sbmlstr)
    is_match = bool(re.search('function', antimony_str))
    if is_match==True:
        raise NotImplementedError('SBML string is too complex for module to convert')
    else:
        return



import configparser
###############################################################################
#Non-Standard Imports
###############################################################################
try:
    from .                   import model_handler    as mh 
    from .                   import settings_handler as sh
    from .setup_sim          import (compile_models, setup_helper, 
                                     string_to_dict, string_to_dict_array, 
                                     string_to_list_string, eval_init_string, 
                                     eval_params_string, eval_tspan_string, 
                                     dict_template, list_template, t_template,
                                     is_analysis_settings)

except:
    import model_handler    as     mh
    import settings_handler as     sh
    from   setup_sim        import (compile_models, setup_helper, 
                                     string_to_dict, string_to_dict_array, 
                                     string_to_list_string, eval_init_string, 
                                     eval_params_string, eval_tspan_string, 
                                     dict_template, list_template,
                                     is_analysis_settings)
    
###############################################################################
#Interfacing with Configparser
###############################################################################
def from_config(filename, sampler='sa'):
    '''Opens a config file and reads the fields/subfields required for setting up 
    the analysis while ignoring the irrelavant ones. Returns a dictionary of the 
    collected information.
    
    :param filename: Name of file to read.
    :type filename: str
    :param sampler: "sa" for simulated annealing; "bh" for scipy's basin hopping;
        "da" for scipy's dual annealing; "de" for scipy's differential evolution; 
        defaults to "sa"
    :type sampler: str, optional
    '''
    config             = configparser.RawConfigParser()
    config.optionxform = lambda option: option
    config_data               = {}
    with open(filename, 'r') as file:
        config.read_file(file)
    
    n = 1
    for section in config.sections():
        if not is_analysis_settings(config, section):
            continue
        
        solver_args      = config[section].get('solver_args')
        fixed_parameters = config[section].get('fixed_parameters')
        parameter_bounds = config[section].get('parameter_bounds')
        priors           = config[section].get('priors')
        step_size        = config[section].get('step_size')
        init             = config[section].get('init')
        tspan            = config[section].get('tspan')
        sa_args          = config[section].get('sa_args')
        de_args          = config[section].get('de_args')
        bh_args          = config[section].get('bh_args')
        da_args          = config[section].get('da_args')
        op_args          = config[section].get('op_args')
        guess            = config[section].get('guess')
        units            = config[section].get('units')
        
        tspan            = eval_tspan_string(tspan)
        init             = eval_init_string(init)                  if init             else {}
        solver_args      = string_to_dict(solver_args)             if solver_args      else {}
        fixed_parameters = string_to_list_string(fixed_parameters) if fixed_parameters else []
        parameter_bounds = string_to_dict_array(parameter_bounds)  if parameter_bounds else {}
        priors           = string_to_dict(priors)                  if priors           else {}
        step_size        = string_to_dict(step_size)               if step_size        else {}
        sa_args          = string_to_dict(sa_args)                 if sa_args          else {}
        de_args          = string_to_dict(de_args)                 if de_args          else {}
        bh_args          = string_to_dict(bh_args)                 if bh_args          else {}
        da_args          = string_to_dict(da_args)                 if da_args          else {}
        op_args          = string_to_dict(op_args)                 if op_args          else {}
        guess            = eval_params_string(guess)               if guess            else {}
        units            = string_to_dict(units)                   if units            else {}
        
        if sampler == 'sa':
            config_data[n] = {'system_type' : section,     'guess'  : guess,  'priors'           : priors, 
                              'solver_args' : solver_args, 'parameter_bounds' : parameter_bounds,
                              'step_size'   : step_size,   'fixed_parameters' : fixed_parameters,   
                              'units'       : units,       'init'   : init,   'tspan' : tspan, 'sa_args':sa_args
                              }
        
        elif sampler == 'de':
            config_data[n] = {'system_type' : section,     'guess'  : guess,  'priors' : priors, 
                              'solver_args' : solver_args, 'parameter_bounds' : parameter_bounds,  
                              'de_args'     : de_args,     'fixed_parameters' : fixed_parameters,
                              'units'       : units,       'init'   : init,   'tspan' : tspan        
                              }
            
        elif sampler == 'bh':
            config_data[n] = {'system_type' : section,     'guess'  : guess,  'priors' : priors, 
                              'solver_args' : solver_args, 'parameter_bounds' : parameter_bounds,  
                              'step_size'   : step_size,   'fixed_parameters' : fixed_parameters,   
                              'bh_args'     : bh_args,     'units'  : units,  'init'   : init,   'tspan' : tspan  
                              }
        elif sampler == 'da':
            config_data[n] = {'system_type' : section,     'guess'  : guess,  'priors' : priors, 
                              'solver_args' : solver_args, 'parameter_bounds' : parameter_bounds,   
                              'da_args'     : da_args,     'fixed_parameters' : fixed_parameters, 
                              'units'       : units,       'init'   : init,   'tspan' : tspan 
                              }
        
        elif sampler == 'op':
            config_data[n] = {'system_type' : section,     'guess'  : guess,  'priors' : priors, 
                              'solver_args' : solver_args, 'parameter_bounds' : parameter_bounds,   
                              'op_args'     : op_args,     'fixed_parameters' : fixed_parameters, 
                              'units'       : units,       'init'   : init,   'tspan' : tspan 
                              }
        
        else:
            raise Exception(f'Invalid sampler: {sampler}')
        n += 1
 
    return config_data

###############################################################################
#Main Set Up
###############################################################################    
def get_sampler_args(filename, sampler='sa', user_core_models={}):
    '''Reads the config file and adds combines it with core_model data. If you are
    using a core model that is not in the database, you must provide the core_model
    using the core_models argument where the key is the system_type. Returns a 
    dictionary of keyword arguments and the config_data.
    
    :param filename: The name of the file to read
    :type filename: str
    :param sampler: "sa" for simulated annealing; "bh" for scipy's basin hopping;
        "da" for scipy's dual annealing; "de" for scipy's differential evolution; 
        defaults to "sa"
    :type sampler: str, optional
    :param user_core_model: A dictionary of core_models indexed by their system_type.
        core_models already in the database do not need to be specified here.
    :type user_core_model: dict, optional
    '''
    if sampler == 'sa':
        return get_sampler_args_sa(filename, user_core_models=user_core_models)
    elif sampler == 'de':
        return get_sampler_args_de(filename, user_core_models=user_core_models)
    elif sampler == 'bh':
        return get_sampler_args_bh(filename, user_core_models=user_core_models)
    elif sampler == 'da':
        return get_sampler_args_da(filename, user_core_models=user_core_models)
    elif sampler == 'op':
        return get_sampler_args_op(filename, user_core_models=user_core_models)
    else:
        raise Exception('sampler must be sa, de, bh or da. sampler given was ' + str(sampler))
    
def get_sampler_args_sa(filename, user_core_models={}):
    '''Shortcut for setting up simulated annealing for one model
    
    :meta: private
    '''

    config_data, core_models = setup_helper(filename, from_config, user_core_models)
    models, params = compile_models(core_models, config_data)
    
    guess            = {}
    bounds           = {}
    priors           = {}
    step_size        = {}
    blocks           = []
    fixed_parameters = []
    trials           = []
    for key in config_data:
        models[key]['init']                    = config_data[key]['init']
        models[key]['tspan']                   = config_data[key]['tspan']
        models[key]['int_args']['solver_args'] = config_data[key]['solver_args']
        
        temp      = {param + '_' + str(key): float(config_data[key]['guess'][param]) for param in config_data[key]['guess']}
        guess     = {**guess, **temp}
        
        temp      = {param + '_' + str(key): config_data[key]['parameter_bounds'][param] for param in config_data[key]['parameter_bounds']}
        bounds    = {**bounds, **temp} 
        
        temp      = {param + '_' + str(key): config_data[key]['priors'][param] for param in config_data[key]['priors']}
        priors    = {**priors, **temp}
    
        temp      = {param + '_' + str(key): config_data[key]['step_size'][param] for param in config_data[key]['step_size']}
        step_size = {**step_size, **temp}
        
        blocks += [[param + '_' + str(key) for param in block] for block in config_data[key]['sa_args'].get('blocks', [])]
        
        fixed_parameters   += [param + '_' + str(key) for param in config_data[key]['fixed_parameters']]
        fixed_parameters   += [i     + '_' + str(key) for i     in core_models[key-1]['inputs']]
        
        if 'trials' in config_data[key]['sa_args']:
            trials.append(config_data[key]['sa_args']['trials'])
        
    sampler_args = {'data'     : {},
                    'guess'    : guess,
                    'models'   : models, 
                    'fixed_parameters'     : fixed_parameters,
                    'priors'   : priors,
                    'bounds'   : bounds,
                    'blocks'   : blocks,
                    'step_size': step_size,
                    'trials'   : max(trials) if trials else 3000,
                    }
        
    return sampler_args, config_data

def get_sampler_args_de(filename, user_core_models={}):
    '''Shortcut for setting up differential evolution for one model
    
    :meta: private
    '''

    config_data    = from_config(filename, 'de') if type(filename) == str else filename
    core_models    = [mh.quick_search(config_data[key]['system_type']) for key in config_data]
    models, params = compile_models(core_models, config_data)
    
    guess            = {}
    bounds           = {}
    priors           = {}
    fixed_parameters = []
    scipy_args       = {}
    for key in config_data:
        models[key]['init']                    = config_data[key]['init']
        models[key]['tspan']                   = config_data[key]['tspan']
        models[key]['int_args']['solver_args'] = config_data[key]['solver_args']
        
        temp      = {param + '_' + str(key): float(config_data[key]['guess'][param]) for param in config_data[key]['guess']}
        guess     = {**guess, **temp}
        
        temp      = {param + '_' + str(key): config_data[key]['parameter_bounds'][param] for param in config_data[key]['parameter_bounds']}
        bounds    = {**bounds, **temp} 
        
        temp      = {param + '_' + str(key): config_data[key]['priors'][param] for param in config_data[key]['priors']}
        priors    = {**priors, **temp}
    
        fixed_parameters   += [param + '_' + str(key) for param in config_data[key]['fixed_parameters']]
        fixed_parameters   += [i     + '_' + str(key) for i     in core_models[key-1]['inputs']]
        
        scipy_args = config_data[key]['de_args'] if 'de_args' in config_data[key] else scipy_args
        
    sampler_args = {'data'       : {},
                    'guess'      : guess,
                    'models'     : models, 
                    'fixed_parameters'       : fixed_parameters,
                    'priors'     : priors,
                    'bounds'     : bounds,
                    'scipy_args' : scipy_args
                    }
        
    return sampler_args, config_data

def get_sampler_args_bh(filename, user_core_models={}):
    '''Shortcut for setting up differential evolution for one model
    
    :meta: private
    '''

    config_data    = from_config(filename, 'bh') if type(filename) == str else filename
    core_models    = [mh.quick_search(config_data[key]['system_type']) for key in config_data]
    models, params = compile_models(core_models, config_data)
    
    guess            = {}
    bounds           = {}
    priors           = {}
    fixed_parameters = []
    step_size        = {}
    scipy_args       = {}
    for key in config_data:
        models[key]['init']                    = config_data[key]['init']
        models[key]['tspan']                   = config_data[key]['tspan']
        models[key]['int_args']['solver_args'] = config_data[key]['solver_args']
        
        temp      = {param + '_' + str(key): float(config_data[key]['guess'][param]) for param in config_data[key]['guess']}
        guess     = {**guess, **temp}
        
        temp      = {param + '_' + str(key): config_data[key]['parameter_bounds'][param] for param in config_data[key]['parameter_bounds']}
        bounds    = {**bounds, **temp} 
        
        temp      = {param + '_' + str(key): config_data[key]['priors'][param] for param in config_data[key]['priors']}
        priors    = {**priors, **temp}
        
        temp      = {param + '_' + str(key): config_data[key]['step_size'][param] for param in config_data[key]['step_size']}
        step_size = {**step_size, **temp}
        
        fixed_parameters   += [param + '_' + str(key) for param in config_data[key]['fixed_parameters']]
        fixed_parameters   += [i     + '_' + str(key) for i     in core_models[key-1]['inputs']]
        
        scipy_args = config_data[key]['bh_args'] if 'bh_args' in config_data[key] else scipy_args
        
    sampler_args = {'data'               : {},
                    'guess'              : guess,
                    'models'             : models, 
                    'fixed_parameters'               : fixed_parameters,
                    'priors'             : priors,
                    'bounds'             : bounds,
                    'step_size'          : step_size,
                    'scipy_args'         : scipy_args,
                    'step_size_is_ratio' : False
                    }
        
    return sampler_args, config_data

def get_sampler_args_da(filename, user_core_models={}):
    '''Shortcut for setting up differential evolution for one model
    
    :meta: private
    '''

    config_data    = from_config(filename, 'da') if type(filename) == str else filename
    core_models    = [mh.quick_search(config_data[key]['system_type']) for key in config_data]
    models, params = compile_models(core_models, config_data)
    
    guess            = {}
    bounds           = {}
    priors           = {}
    fixed_parameters = []
    scipy_args       = {}

    for key in config_data:
        models[key]['int_args']['solver_args'] = config_data[key]['solver_args']
        
        temp      = {param + '_' + str(key): float(config_data[key]['guess'][param]) for param in config_data[key]['guess']}
        guess     = {**guess, **temp}
        
        temp      = {param + '_' + str(key): config_data[key]['parameter_bounds'][param] for param in config_data[key]['parameter_bounds']}
        bounds    = {**bounds, **temp} 
        
        temp      = {param + '_' + str(key): config_data[key]['priors'][param] for param in config_data[key]['priors']}
        priors    = {**priors, **temp}
        
        fixed_parameters   += [param + '_' + str(key) for param in config_data[key]['fixed_parameters']]
        fixed_parameters   += [i     + '_' + str(key) for i     in core_models[key-1]['inputs']]
        
        scipy_args = config_data[key]['da_args'] if 'da_args' in config_data[key] else scipy_args
        
    sampler_args = {'data'               : {},
                    'guess'              : guess,
                    'models'             : models, 
                    'fixed_parameters'               : fixed_parameters,
                    'priors'             : priors,
                    'bounds'             : bounds,
                    'scipy_args'         : scipy_args,
                    }
        
    return sampler_args, config_data

def get_sampler_args_op(filename, user_core_models={}):
    '''Shortcut for setting up differential evolution for one model
    
    :meta: private
    '''

    config_data    = from_config(filename, 'op') if type(filename) == str else filename
    core_models    = [mh.quick_search(config_data[key]['system_type']) for key in config_data]
    models, params = compile_models(core_models, config_data)
    
    guess            = {}
    bounds           = {}
    priors           = {}
    fixed_parameters = []
    scipy_args       = {}

    for key in config_data:
        models[key]['int_args']['solver_args'] = config_data[key]['solver_args']
        
        temp      = {param + '_' + str(key): float(config_data[key]['guess'][param]) for param in config_data[key]['guess']}
        guess     = {**guess, **temp}
        
        temp      = {param + '_' + str(key): config_data[key]['parameter_bounds'][param] for param in config_data[key]['parameter_bounds']}
        bounds    = {**bounds, **temp} 
        
        temp      = {param + '_' + str(key): config_data[key]['priors'][param] for param in config_data[key]['priors']}
        priors    = {**priors, **temp}
        
        fixed_parameters   += [param + '_' + str(key) for param in config_data[key]['fixed_parameters']]
        fixed_parameters   += [i     + '_' + str(key) for i     in core_models[key-1]['inputs']]
        
        scipy_args = config_data[key]['op_args'] if 'op_args' in config_data[key] else scipy_args
        
    sampler_args = {'data'               : {},
                    'guess'              : guess,
                    'models'             : models, 
                    'fixed_parameters'   : fixed_parameters,
                    'priors'             : priors,
                    'bounds'             : bounds,
                    'scipy_args'         : scipy_args,
                    }
        
    return sampler_args, config_data

###############################################################################
#Template Generation
###############################################################################    
def make_settings_template(system_types_settings_names, filename='', user_core_models={}):
    '''Writes settings to a config file. If you are using a core_model that is 
    not in the database, you must provide the core_model using the core_models 
    argument where the key is the system_type.Returns the code as a string.
    
    :param system_types_settings_names: Pairs of tuples containing (system_type, settings_name)
    :type system_types_settings_names: list or tuple
    :param filename: The name of the file to write to
    :type filename: str, optional
    :param user_core_model: A dictionary of core_models indexed by their system_type.
        core_models already in the database do not need to be specified here.
    :type user_core_model: dict, optional
    '''
    result = ''
    
    system_types_settings_names1 = [system_types_settings_names] if type(system_types_settings_names) == str else system_types_settings_names
    
    for pair in system_types_settings_names1:
        try:
            system_type, settings_name = pair
        except:
            system_type, settings_name = pair, '__default__'

        core_model     = user_core_models[system_type] if system_type in user_core_models else mh.quick_search(system_type)
        parameters     = core_model['parameters'] + core_model['inputs']
        states         = core_model['states']
        section_header = '[' + core_model['system_type'] + ']'
        
        settings = {'system_type'      : system_type, 
                    'settings_name'    : settings_name,
                    'parameters'       : {},
                    'units'            : {},
                    'init'             : {},  
                    'priors'           : {}, 
                    'parameter_bounds' : {},
                    'tspan'            : [],
                    'fixed_parameters' : [],
                    'solver_args'      : {'rtol'   : 1.49012e-8,
                                          'atol'   : 1.49012e-8,
                                          'tcrit'  : [],
                                          'h0'     : 0.0,
                                          'hmax'   : 0.0,
                                          'hmin'   : 0.0,
                                          'mxstep' : 0
                                          },
                    }
        
        if settings_name:
            settings = sh.quick_search(system_type, settings_name, skip_constructor=True)
        
        longest          = len(max(parameters,               key=len))    
        longest_         = len(max(settings['solver_args'],  key=len))  
        solver_args      = settings['solver_args'].keys()
        param_values     = dict_template('parameter_values', parameters, longest, settings['parameters'])
        parameter_bounds = dict_template('parameter_bounds', parameters, longest, settings['parameters'])
        priors           = dict_template('priors',           parameters, longest, settings['parameters'])
        units_values     = dict_template('units',            parameters, longest, settings['units'], default='')
        fixed_parameters = list_template('fixed_parameters', settings['fixed_parameters'])
        solver_args      = dict_template('solver_args',      solver_args, longest_, settings['solver_args'])
   
        model_id         = '#id = ' + str(core_model['id'])
        model_equations  = '#equations = \n' + '\n'.join(['#\t' + line for line in core_model['equations'] ])
        section_header   = '\n'.join([section_header, model_id, model_equations])
        
        result += '\n\n'.join([section_header, param_values, units_values, parameter_bounds, priors, fixed_parameters, solver_args])
        
    if filename:
        with open(filename, 'w') as file:
            file.write(result)
    return result






from   sympy import symbols
from   sympy.matrices import Matrix, zeros
from   math           import ceil
import string
import sympy as sp

from time import time

nt = 10

def t(*funcs):
    global nt
    for func in funcs:
        start = time()
        for i in range(nt):
            func()
        print(time()-start)

def timethis(message):
    def wrapper(func):
        def helper(*args, **kwargs):
            start  = time()
            result = func(*args, **kwargs)
            print(message, time()-start)
            return result
        return helper
    return wrapper

def format_to_matlab(mat_row):
    '''
    For backend testing only. Do not run.
    '''
    s = str(mat_row)
    s = s[7:len(s)-1]
    s = s[1:len(s)-1]
    s = s.replace('**', '^')

    with open('matlab_comparison.txt', 'w') as file:
        file.write(s)
    return s

###############################################################################
#Globals
###############################################################################
verbose = False

###############################################################################
#Building OI
###############################################################################
@timethis('Time for adding one Lie derivative:')
def add_block(curr_block, block_num, input_der, f_aug, x_aug, last_L):
    '''
    Find the new Lie derivative and create a new block for Oi
    '''
    global verbose
    dummy      = symbols(['dummy'])[0]
    L_h        = curr_block*f_aug
    extra      = zeros(curr_block.shape[0], 1)
    if input_der:
        #Modification for accounting for inputs
        for i in range(block_num-1):
            if i < input_der.shape[1] - 1:
                lo_u   = input_der[:,i]
                hi_u   = input_der[:,i+1]
                lo_u   = lo_u.subs(0, dummy)
                
                if all([True if i == 0 or i == dummy else False for i in lo_u]):
                    continue
                
                extra_ = last_L.jacobian(Matrix(lo_u)) *hi_u
                extra += extra_
            else:
                break

    last_L     = L_h + extra
    
    new_block  = last_L.jacobian(x_aug)
    if verbose:
        print(new_block)

    return new_block, last_L

def get_nd(h, x, p):
    return ceil((len(x) +len(p) -len(h))/len(h))

def build_input_der(u, nd):
    u_var = list(u.keys())
    u_nzd = list(u.values())
    
    input_der = []
    for i in range(len(u)): 
        if u_nzd[i] == 'inf':
            new_sym = [u_var[i]] + symbols([u_var[i].name+'_d'+str(ii+1) for ii in range(nd-1)])
        else:
            nu      = min(u_nzd[i], nd-1) 
            new_sym = [u_var[i]] + symbols([u_var[i].name+'_d'+str(ii+1) for ii in range(nu)]) + [0]*(nd-1-nu)
        input_der.append(new_sym)

    return Matrix(input_der)

@timethis('Time for building Oi:')
def build_OI(x, p, f, h, u, nd, ics={}):
    x_aug = Matrix(list(x) + list(p)).T
    f_aug = Matrix(list(f) + [0]*len(p))

    Oi        = zeros(rows=len(h)*(nd+1), cols=len(x_aug))
    input_der = build_input_der(u, nd)

    print('Building Oi with '+ str(nd) + ' derivatives and size', Oi.shape)
    
    curr_block    = h.T.jacobian(x_aug)
    Oi[:len(h),:] = curr_block
    last_L        = h
    for block_num in range(1, nd+1):
        curr_block, last_L       = add_block(curr_block, block_num, input_der, f_aug, x_aug, last_L)

        if ics:
            curr_block = curr_block.subs(ics.items())
        
        start                    = block_num*len(h)
        Oi[start:start+len(h),:] = curr_block

    return Oi, input_der, x_aug, f_aug, last_L

def extend_Oi(Oi, input_der, nd, x_aug, f_aug, x, p, f, h, u, last_L):
    print('Extending Oi')
    new_Oi = Matrix(Oi.tolist() + [[0]*len(x_aug)]*len(h))
    
    curr_block    = Oi[-len(h):,:]
    block_num     = nd + 1
    new_input_der = build_input_der(u, block_num)

    new_Oi[-len(h):,:], last_L = add_block(curr_block, block_num, new_input_der, f_aug, x_aug, last_L)

    return new_Oi, block_num, last_L

###############################################################################
#Elimination and Recalculation
###############################################################################
def elim_recalc(Mr, rank, x_aug, h, skip):
    '''
    Assumes initial conditions have been substituted already.
    '''
    print('Elim-recalc for matrix with size', Mr.shape)
    found = []
    
    #Eliminate var
    for i in range(len(x_aug)):
        var = x_aug[i]
        
        if var in skip or var in h or i not in rank:
            continue
        Mr_        = Mr.copy()
        Mr_.col_del(i)
        Mr_, rank_ = Mr_.rref()

        if len(rank) != len(rank_):
            found.append(var)
    print('Found', found)
    return found

@timethis('Time for Row Reduction:')
def check_rank(Oi, ics):
    print('Checking rank for matrix with size', Oi.shape)
    Mr       = Oi.subs(ics.items())
    if 'inf' in  Mr.__str__():
        raise Exception('inf found in Ms. This could be due to a badly chosen initial value.')
    Mr, rank = Mr.rref()
    print('Rank', len(rank))
    return Mr, rank

###############################################################################
#Core Algorithm
############################################################################### 
def strike_group(x, p, f, h, u, ics):
    '''
    Core algorithm. Edit with extreme caution.
    Builds Oi and checks rank.
    Adds a Lie derivative if necessary and checks rank again.
    Note that the current implementation always uses decomp=False.
    '''
    nd = get_nd(h, x, p)
    Oi, input_der, x_aug, f_aug, last_L = build_OI(x, p, f, h, u, nd, ics)
        
    Mr, rank = check_rank(Oi, ics)

    if len(rank) == len(x_aug):
        found = [i for i in x_aug if i not in h]
        print('Found', found)
        
    else:
        found  = elim_recalc(Mr, rank, x_aug, h, [])
        
        while True:
            Oi, nd, last_L = extend_Oi(Oi.subs(ics.items()), input_der, nd, x_aug, f_aug, x, p, f, h, u, last_L)

            Oi     = Matrix(Mr.tolist() + Oi[-len(h):,:].tolist())
            
            Mr, rank = check_rank(Oi, ics)
            
            if len(rank) == len(x_aug):
                new_found = [i for i in x_aug if i not in h and i not in found]
                found     = [i for i in x_aug if i not in h]
                print('Found', new_found)
                break
            else:
                found_= elim_recalc(Mr, rank, x_aug, h, found)

                if found_:
                    found += found_
                    
                else:
                    break
    return found

###############################################################################
#Decomposition
###############################################################################
def make_group(h, x, p, u, f, ics, group, found):
    '''
    Creates submodels(groups).
    States and parameters in found will not be included in
    the unknowns for analysis.
    '''
    
    xg    = Matrix([x[i] for i in range(len(x)) if x[i] in group])
    fg    = Matrix([f[i] for i in range(len(f)) if x[i] in xg])
    
    terms = get_terms(fg)
    pg    = Matrix([i  for i in p   if  str(i) in terms and  i not in found] +
                   [i  for i in x   if  str(i) in terms and  i not in found and i not in xg and i not in h])
    hg    = Matrix([i  for i in x   if  i      in xg    and (i     in found or i in h)])
    ug    = {i: u[i]   for i in u   if  str(i) in terms}
    icsg  = ics.copy()#{i: ics[i] for i in ics if  i      in xg}

    return hg, xg, pg, ug, fg, icsg

def get_terms(f):
    '''
    Supporting function for make_group. Extracts all terms in an expression.
    '''
    punctuation = string.punctuation.replace('_', '')
    fs = str([i for i in f])
    fs = fs[1:len(fs)-1].replace(' ', '')
    tr = str.maketrans(punctuation, ','*len(punctuation)) 
    fs = fs.translate(tr)
    fs = [i for i in fs.split(',') if i]

    return set(fs)

###############################################################################
#Main Algorithm
###############################################################################
def iterative_decomp(x, p, f, h, u, ics, decomp=[]):
    '''
    Iterates across groups and updates accordingly.
    '''
    found = []
    
    for group in decomp:
        start = time()
        hg, xg, pg, ug, fg, icsg = make_group(h, x, p, u, f, ics, group, found)
        print('States : ', list(xg))
        print('Outputs: ', list(hg))
        print('Params : ', list(pg))
        print('Inputs : ', ug)
        print('ICS    : ', icsg)
        print('Equations:\n'+'\n'.join([str(i) for i in fg]))

        found_ = strike_group(xg, pg, fg, hg, ug, icsg)
        print(time()-start)
        print('...................................................')
        
        found += found_
    return list(h) + found

@timethis('Total Time:')
def strike_goldd(h, x, p, u, f, ics, decomp=[]):
    '''
    Wrap the algorithm and organize the results.
    Also checks input arguments for errors.
    '''
    if len(f) != len(x):
        raise Exception('Number of equations not equal to number of states! '+
                        'Remember to mark on non-differential expressions with "!" if you are using a mark-up file.')
    
    if len(ics) != len(x):
        raise Exception('Number of initial conditions not equal to number of states!')

    decomp1    = decomp if decomp else [list(x)]
    found      = iterative_decomp(x, p, f, h, u, ics, decomp=decomp1)
    x_aug_dict = {i: True if i in found else False for i in list(x)+list(p)}
        
    return x_aug_dict

def analyze_sg_args(sg_args, dst={}):
    result = dst if dst else {}
    for key in sg_args:
        result[key] = strike_goldd(**sg_args[key])
    
    return result
        
if __name__ == '__main__':
    pass
##    m1, p1 = symbols(['m1', 'p1'])
##    synm1, degm, synp1, degp = symbols(['synm', 'degm', 'synp', 'degp'])
##    g1, g2, g3, g4 = symbols(['g1', 'g2', 'g3', 'g4'])
##    u1 = symbols(['u1'])[0]
##
##    dm1 = synm1 - degm*m1 - u1*g1*g2*g3*g4
##    dp1 = synp1*m1 - degp*p1 + u1
##    
##    x = Matrix([m1, p1])
##    h = Matrix([p1])
##    p = Matrix([synm1, degm, synp1, degp, g1, g2, g3, g4])
##    f = Matrix([dm1, dp1])
##    u = {u1: 0}
##    
##    x_aug = Matrix(list(x) + list(p)).T
##    f_aug = Matrix(list(f) + [0]*len(p))
##
##    nd        = get_nd(h, x, p)
##    Oi        = zeros(rows=len(h)*(nd+1), cols=len(x_aug))
##    input_der = build_input_der(u, nd)
##
##    print('Building Oi with '+ str(nd) + ' derivatives.')
##    
##    curr_block  = h.T.jacobian(x_aug)
##    Oi[:len(h),:] = curr_block
##    
##    for block_num in range(1, nd+1):
##        curr_block               = add_block(curr_block, block_num, input_der, f_aug, x_aug)
##        start                    = block_num*len(h)
##        Oi[start:start+len(h),:] = curr_block



##    a, n1, r3, g1, g2, k1, k2, k3, k4, k5, k6, k7, m1, m2, m3, m4, m5, m6, m7, n2, p1, p2, p3, q1, q2, r1, r2, r4 = symbols(['a', 'n1', 'r3', 'g1', 'g2', 'k1', 'k2', 'k3', 'k4', 'k5', 'k6', 'k7', 'm1', 'm2', 'm3', 'm4', 'm5', 'm6', 'm7', 'n2', 'p1', 'p2', 'p3', 'q1', 'q2', 'r1', 'r2', 'r4'])
##    x1, x2, x3, x4, x5, x6, x7 = symbols(['x1', 'x2', 'x3', 'x4', 'x5', 'x6', 'x7'])
##    u1 = symbols(['u1'])[0]
##
##    dx1 = n1*x6**a/(g1**a+x6**a) -m1*x1/(k1+x1) +q1*x7*u1
##    dx1 = dx1.subs(a, 1)
##    dx2 = p1*x1 -r1*x2 +r2*x3    -m2*x2/(k2+x2)    
##    dx3 = r1*x2 -r2*x3           -m3*x3/(k3+x3)
##    dx4 = n2*g2**2/(g2**2+x3**2) -m4*x4/(k4+x4)
##    dx5 = p2*x4 -r3*x5 +r4*x6    -m5*x5/(k5+x5)
##    dx6 = r3*x5 -r4*x6           -m6*x6/(k6+x6)
##    dx7 = p3                     -m7*x7/(k7+x7) -(p3+q2*x7)*u1
##
##    x = Matrix([x1, x2, x3, x4, x5, x6, x7])
##    p = Matrix([n1, r3, g1, g2, m1, m2, m3, m4, m5, m6, m7, n2, p1, p2, p3, q1, q2, r1, r2, r4])
##    f = Matrix([dx1, dx2, dx3, dx4, dx5, dx6, dx7])
##    h = Matrix([x1, x2, x3, x4])
##    u = {u1: 0}
##    ics = {state: 0 for state in x}
##
##    found = []
##    
##    hg, xg, pg, ug, fg, icsg = make_group(h, x, p, u, f, ics, [x1, x7], [])
##    decomp = [[x1, x4], [x2, x3], [x3, x4], [x1, x7]]
##    decomp = [[x1, x7]]
####    found = iterative_decomp(x, p, f, h, u, ics, decomp=decomp)
##
##    nd = get_nd(hg, xg, pg)
##    Oi, input_der, x_aug, f_aug = build_OI(xg, pg, fg, hg, ug, 9)


    
##    print(found)
##    
##    h = Matrix([i for i in x if i in h or i in found])
##    p = Matrix([i for i in p if i not in found])
##
##    found = iterative_decomp(x, p, f, h, u, ics, decomp=[[x2, x3]])
##
##    
##    answer = set([x1, x2, x3, x4, m4, n2, x6, x7, x2, p3, q1, m1, m7, q2])
    

############################################################
############################################################
##    found = []
##    
##    hg, xg, pg, ug, fg, icsg = make_group(h, x, p, u, f, ics, [x1, x4], found)
##    
##    Oi, input_der, nd, x_aug, f_aug = build_OI(xg, pg, fg, hg, ug)




import numpy as np
from numpy import log   as ln
from numpy import log10 as log
from numpy import exp
def model_Test_Model_Cell_Cycle(y, t, params):
	EmptySet = y[0]
	u        = y[1]
	z        = y[2]
	v        = y[3]

	kappa   = params[0]
	k6      = params[1]
	k4      = params[2]
	k4prime = params[3]
	alpha   = params[4]

	alpha= k4prime/k4
	z = v - u
	du= k4*(v - u)*(alpha + u**2) - k6*u
	dv= kappa - k6*u
	dEmptySet= +(k6*u) -(kappa)
	du= +(k4*z*(k4prime/k4 + u**2)) -(k6*u)
	dz= +(kappa) -(k4*z*(k4prime/k4 + u**2))

	return np.array([dEmptySet, du, dz, dv])

EmptySet,u,z,v,kappa,k6,k4,k4prime,alpha= np.random.rand(9)*10

EmptySet,u,z,v,kappa,k6,k4,k4prime,alpha= list(map(float, [EmptySet,u,z,v,kappa,k6,k4,k4prime,alpha]))

y = [EmptySet,u,z,v]

t = 0
dt = 1e-3

params = [kappa,k6,k4,k4prime,alpha]

y = y + dt*model_Test_Model_Cell_Cycle(y, t, params)

y = y + dt*model_Test_Model_Cell_Cycle(y, t, params)


import numpy  as np
import pandas as pd
from   scipy.stats import norm

###############################################################################
#Non-Standard Imports
###############################################################################
from . import curvefitting     as cf
   
###############################################################################
#ic Calculation
###############################################################################    
def calculate_ic(data, models, params, priors={}, ictype='AIC', alpha=0.2):
    '''Calculates ic of model calculating posterior and than applying ic 
    formula. Returns a DataFrame of the ic Values.

    Parameters
    ----------
    data : dict
        Curve-fitting data.
    models : dict
        The models used for curve-fitting.
    params : pandas.DataFrame
        The parameters with which to integrate the models with.
    priors : dict, optional
        Priors if any. The default is {}.
    ictype : {'AIC', 'BIC', 'CAIC'}
        The information criterion used.
    alpha : float, optional
        Controls the ratio of AIC/BIC when using CAIC. Ignored when ictype is not 
        'CAIC'. The default is 0.2.

    Returns
    -------
    table : pandas.DataFrame
        A DataFrame with ic values.
    '''
    ic = {}
    
    #Standardize format
    try:
        params1 = pd.DataFrame(params)
    except:
        params1 = pd.DataFrame([params])
        
    #Set up arguments
    posterior_args              = get_posterior_args(data, models, params1, priors)
    n_points, n_params, samples = get_ic_args(data, models)
    
    ic = []  

    for name, row in params1.iterrows():
        log_posterior = get_posterior_for_each_model(params_series=row, **posterior_args) 
        
        if ictype == 'AIC':
            model_ic = n_points*np.log(-log_posterior/n_points) + 2*n_params
        elif ictype == 'BIC':
            model_ic = n_points*np.log(-log_posterior/n_points) + np.log(sample_size)*n_params
        elif ictype == 'CAIC':
            aic = n_points*np.log(-log_posterior/n_points) + 2*n_params
            bic = n_points*np.log(-log_posterior/n_points) + np.log(sample_size)*n_params
            model_ic = alpha*aic-(1-alpha)*bic
        else:
            raise ValueError(f'Invalid value given for ictype: {ictype}')
        ic.append(model_ic)
        
    ic = pd.DataFrame(ic, columns=models.keys(), index=params1.index)
    
    table         = pd.DataFrame(ic.stack().reset_index())
    table.columns = ['row', 'model_num', 'ic value']
    return table

def get_ic_args(data, models):
    '''
    Supporting function for calculate_ic. Do not run.
    
    :meta private:
    '''
    n_points = {}
    n_params = []
    sample   = [] 
    for model_num in data:
        n_points[model_num] = 0
        n_params.append( len(models[model_num]['params']) )
        temp = 0
        for state in data[model_num]:
            for scenario in data[model_num][state]:
                if scenario < 1:
                    continue
                else:
                    n_points[model_num] += len(data[model_num][state][scenario])
                    temp += 1
        sample.append(temp)
        
    return np.array(list(n_points.values())), np.array(n_params), np.array(sample)

def get_posterior_args(data, models, params, priors):
    '''
    Supporting function for calculate_ic. Do not run.
    
    :meta private:
    '''
    t_indices    = cf.get_t_indices(data, models)

    prior_distribution = {model_num: np.array([priors[param] for param in models[model_num]['params'] if param in priors]) for model_num in models}
    prior_distribution = {model_num: norm(prior_distribution[model_num][:,0], prior_distribution[model_num][:,1]) for model_num in models if len(prior_distribution[model_num])}
    prior_index        = {model_num: [param for param in models[model_num]['params'] if param in priors] for model_num in models}
    
    result        =  {'data'               : data, 
                      'models'             : models, 
                      't_indices'          : t_indices, 
                      'prior_distribution' : prior_distribution,
                      'prior_index'        : prior_index
                      }
    return result        

def get_posterior_for_each_model(data, models, params_series, prior_distribution, prior_index, t_indices={}):
    '''
    Supporting function for calculate_ic. Do not run.
    
    :meta private:
    '''
    log_posterior = []
    for model_num in data:
        
        #Log likelihood
        #SSE negated due to definition of log likelihood
        params_          = params_series[models[model_num]['params']].values
        model_likelihood = -cf.get_SSE_dataset(model_num, data, models, params_, t_indices)
        
        #Log prior
        if len(prior_index[model_num]):
            params__    = params_series[prior_index[model_num]]
            model_prior = np.sum(prior_distribution[model_num].pdf(params__))
        else:
            model_prior = 0
        
        model_posterior = model_likelihood + model_prior
        
        log_posterior.append(model_posterior)
    return np.array(log_posterior)
    
    
def normalize_ic(ic_values):
    '''
    Accepts a list-like array of ic values and returns a normalized array 
    where the lowest value is zero.
    '''
    return np.array(ic_values) - np.min(ic_values)

###############################################################################
#ic Tables/DataFrames
###############################################################################
def rank_ic(ic_dataframe, ic_column_name='ic value', inplace=True):
    '''
    Accepts a dataframe where each row corresponds to one model
    ic_dataframe cannot have any columns named 'Evidence'.
    '''
    dic     = normalize_ic(ic_dataframe[ic_column_name])
    evidence = ['Substantial support' if dic[i] < 2 else 'Weak support' if dic[i] < 10 else 'No support' for i in range(len(dic))] 
    
    loc  = len(ic_dataframe.columns) 
    
    ranked_table = ic_dataframe if inplace else ic_dataframe.copy()
    ranked_table.insert(loc    = loc,
                        column ='d(' + ic_column_name + ')',
                        value  = dic
                        ) 
    ranked_table.insert(loc    = loc,
                        column ='Evidence',
                        value  = evidence
                        ) 
    
    ranked_table = ranked_table.sort_values(ic_column_name)
    return ranked_table
    



import importlib
import os
import os.path    as osp
from   pathlib    import Path

bmh, bsh = None, None
    
###############################################################################
#Globals
###############################################################################
__dir__ = osp.dirname(osp.abspath(__file__))

###############################################################################
#Reset
###############################################################################
def reset_MBase(*other_files):
    '''
    Provides a safe way to reset MBase. Close all connections to BMSS first. 
    Restart the kernel if necessary.
    '''
    global bmh
    global bsh
    
    to_del = ['MBase.db', 'UBase.db'] + list(other_files)
    
    print(to_del)
    for filename in os.listdir():
        if filename in to_del:
            os.remove(filename)
            msg = 'Deleted {}'.format(filename)
            print(msg)
            
    for filename in os.listdir('model_functions'):
        if '__' not in filename:
            os.remove('model_functions/' + filename)
            msg = 'Removed {} from model_functions.'.format(filename)
            print(msg)
    
    bmh = importlib.import_module('model_handler')
    bsh = importlib.import_module('settings_handler')
    
    #For loading markup files in markup to MBase
    markup_directory = Path(os.getcwd()) /'BMSS_markup'
    for f in os.listdir(markup_directory):
        print('Reading', f)
        filename   = markup_directory/f
        core_model = bmh.from_config(filename)
        
        bmh.backend_add_to_database(core_model, database=bmh.MBase)
        bmh.model_to_code(core_model)
        bsh.backend_config_to_database(filename, bmh.MBase)
    
if __name__ == '__main__':
    '''
    RESTART THE KERNEL OR THIS WILL THROW AN ERROR!
    '''
    reset_MBase()


import os
import pandas  as     pd
import yaml    as     ya
from   io      import StringIO
from   pathlib import Path

###############################################################################
#Non-Standard Imports
###############################################################################
try:
    from . import model_handler as mh
except:
    import model_handler as mh

###############################################################################
#Globals
###############################################################################
ia_results_folder = Path(os.path.dirname(__file__)) / 'ia_results'

###############################################################################
#IA Result Handling
###############################################################################
def export_sg_results(sg_results, variables, config_data, user_core_models={}, filename='sg_results.yaml'):
    '''
    Exports the results as a human-readable .yaml file
    '''
    yaml_dict = {}
    
    for key in sg_results:
        system_type = config_data[key].get('system_type')
        core_model  = user_core_models.get(system_type)

        if not core_model:
            core_model = mh.quick_search(system_type)
            
        fixed_parameters = list(config_data[key]['parameter_values'].keys())
        
        nested_dict = make_new_nested_dict(sg_result        = sg_results[key],
                                           model_variables  = variables[key],
                                           core_model       = core_model,
                                           fixed_parameters = fixed_parameters,
                                           **config_data[key]
                                           )
        nested_dict    = {**{'system_type' : system_type}, **nested_dict}
        yaml_dict[key] = nested_dict
    
    outfile = filename if filename[-5:] == '.yaml' else filename + '.yaml'
    
    
    try:
        with open(outfile, 'w') as file:
            ya.dump(yaml_dict, file, sort_keys=False)
    except FileNotFoundError:#user wants to use relative path
        cwd        = Path(os.getcwd())
        rel_path   = Path(outfile)
        folder     = rel_path.parent
        stem       = rel_path.stem 
        ext        = rel_path.suffix
        outfile    = stem + ext if ext == '.yaml' else stem + ext + '.yaml'
        
        os.mkdir(folder)
        
        full_path = cwd / folder / outfile
        
        with open(full_path, 'w') as file:
            ya.dump(yaml_dict, file, sort_keys=False)
        
    except Exception as e:
        raise e
        
    return yaml_dict
    
def dump_sg_results(sg_results, variables, config_data, user_core_models={}, save=False):
    '''
    Adds sg_results for each core_model to the csv file indexed under core_model['ia']
    where the system_type of each core_model is specified in config_data.
    
    If the core_model is not in the database, you must provide your own core_model 
    in user_core_models and index it under its system_type.
    '''
    new_rows = {}
    for key in sg_results:
        system_type = config_data[key].get('system_type')
        core_model  = user_core_models.get(system_type)

        if not core_model:
            core_model = mh.quick_search(system_type)
        
        fixed_parameters = list(config_data[key]['parameter_values'].keys())
        
        new_row = update_core_model(sg_result        = sg_results[key],
                                    model_variables  = variables[key],
                                    core_model       = core_model,
                                    save             = save,
                                    fixed_parameters = fixed_parameters,
                                    **config_data[key]
                                    )
        
        new_rows[key] = new_row
    
    return new_rows

###############################################################################
#Row Generation
###############################################################################
def write_new_row_to_file(sg_result, model_variables, init, input_conditions, fixed_parameters,  measured_states, system_type, core_model, pd_args={}, local=False, **kwargs):
    '''
    Generates a new row by calling make_new_row and updates the file indexed 
    under core_model['ia'] with replacement to prevent duplication. 
    
    The file will be created if it does not exist. model_handler.quick_search 
    is called if the core_model is not provided. 
    '''
    global ia_results_folder
        
    new_row = make_new_row(sg_result, model_variables, init, input_conditions, fixed_parameters,  measured_states, core_model, **kwargs)
    
    
    filename = core_model['ia']
    

    if filename:
        location  = os.getcwd() if local else ia_results_folder
        directory = filename    if local else ia_results_folder / filename 
        
        if filename in os.listdir(location):
            if 'header' not in pd_args:
                pd_args['header'] = [0, 1]
                
            df = pd.read_csv(directory, **pd_args)
            s  = pd.Series(new_row)
            
            if not all(df.columns == s.index):
                raise Exception('Keys in new row do not match csv headers. csv headers: ' + str(df.columns) )
            if len(df.columns) != len(s.index):
                raise Exception('Keys in new row do not match csv headers. csv headers: ' + str(df.columns) )
                
            #Check if inputs are duplicated
            try:
                row_num          = df['known'][s['known'] == df['known']].index[0]
                df.iloc[row_num] = s
                print('Updated row ' + str(row_num) + ' in ' + filename)
            except:
                df = df.append(s, ignore_index=True)
                print('Added row to ' + filename)
        
        else:
            df = pd.DataFrame([new_row], columns=pd.MultiIndex.from_tuples(new_row.keys()))
            df.to_csv(directory, index=False)
            print('Created ' + str(directory) + ' to store strike-goldd results.')
        
    else:
        print('No filename in core_model["ia"]. No files were written.')
    return new_row

def update_core_model(sg_result, model_variables, init, input_conditions, fixed_parameters,  measured_states, system_type, core_model, pd_args={}, save=False, **kwargs):
    '''
    Generates a new row by calling make_new_row and updates the database with replacement 
    if save is True. 
    '''
        
    new_row = make_new_row(sg_result, model_variables, init, input_conditions, fixed_parameters,  measured_states, core_model, **kwargs)
    mh.update_ia(core_model, new_row, save=save)
    
    return new_row

def make_new_row(sg_result, model_variables, init, input_conditions, fixed_parameters,  measured_states, core_model, **kwargs):
    '''
    Generates a row in dictionary form that can be appended to a DataFrame. Each 
    column is a tuple where the first value is "known" or "unknown" and the second
    value is a string variable that corresponds to a state, parameter of input in a
    core_model. Ignores kwargs.
    '''
    new_row = make_new_row_template(core_model, init, input_conditions, fixed_parameters,  measured_states, **kwargs)
    new_row = update_row_with_sg_result(new_row, sg_result, model_variables)
    
    return new_row

def make_new_nested_dict(sg_result, model_variables, init, input_conditions, fixed_parameters,  measured_states, core_model, **kwargs):
    '''
    Generates a nested dict from the return value of make_new_row
    '''
    new_row = make_new_row(sg_result, model_variables, init, input_conditions, fixed_parameters,  measured_states, core_model, **kwargs)
    
    nested_dict = {}
    
    for key, value in new_row.items():
        k1, k2                             = key
        nested_dict.setdefault(k1, {})[k2] = value
    
    return nested_dict

###############################################################################
#Supporting Functions
###############################################################################
def update_row_with_sg_result(row, sg_result, model_variables):
    '''
    Updates the row with information from sg_result and model_variables where
    sg_result is the output from the function strike_goldd and model_variables
    is a dict that maps a str variable in row to a symbolic variable in sg_result.
    '''
    for variable, symbolic in model_variables.items():
        column = 'output', variable
        
        if symbolic in sg_result:
            row[column] = int(sg_result[symbolic])

    return row

def make_new_row_template(core_model, init, input_conditions, fixed_parameters,  measured_states, **kwargs):
    '''
    Creates a dictionary that can be used as a template.
    '''
    new_row    = {}
    states     = core_model['states']
    parameters = core_model['parameters']
    inputs     = core_model['inputs']
    init_      = dict(zip(states, init[1]))

    for column in make_columns(core_model):
        group, variable = column
        
        if group == 'input':
            if check_in_lists(variable, states, measured_states):
                new_row[column] = 1
            elif check_in_lists(variable, parameters, fixed_parameters):
                new_row[column] = 1
            elif variable in inputs:
                new_row[column] = input_conditions[variable]
            else:
                new_row[column] = 0
        elif group == 'init':
            new_row[column] = float(init_[variable])
        else:
            new_row[column] = 1

    return new_row

def check_in_lists(variable, lst1, lst2, negate=False):
    '''
    Supporting function for make_new_row. Do not run.
    
    :meta private:
    '''
    if negate:
        return variable in lst1 and variable not in lst2
    else:
        return variable in lst1 and variable in lst2
    
def make_columns(core_model):
    '''
    Generates column tuples using states and parameters. Inputs are not included
    since 
    '''
    
    second = core_model['states'] + core_model['parameters'] 
    
    left  = [('init', s) for s in core_model['states']] + [('input', s) for s in second + core_model['inputs']]
    right = [('output', s) for s in second]
    return left + right





import configparser
import importlib
import numpy as np
import pandas as pd

###############################################################################
#Non-Standard Import
###############################################################################
try:
    from . import model_handler    as mh
    from . import settings_handler as sh
    from .utils_settings import *
except:
    import model_handler    as mh
    import settings_handler as sh
    from   utils_settings   import *

###############################################################################
#Interfacing with Configparser
###############################################################################    
def from_config(filename):
    '''Opens a config file and reads the fields/subfields required for setting up 
    the analysis while ignoring the irrelavant ones. Returns a dictionary of the 
    collected information.
    
    :param filename: Name of file to read.
    :type filename: str
    '''
    config             = configparser.RawConfigParser()
    config.optionxform = lambda option: option
    config_data               = {}
    with open(filename, 'r') as file:
        config.read_file(file)
    
    n = 1
    for section in config.sections():
        if not is_analysis_settings(config, section):
            continue
        
        init        = config[section]['init']
        params      = config[section]['parameter_values']
        tspan       = config[section]['tspan']
        solver_args = config[section].get('solver_args')
        
        init        = eval_init_string(init)
        tspan       = eval_tspan_string(tspan)
        params      = eval_params_string(params)
        solver_args = string_to_dict(solver_args) if solver_args else {}
            
        config_data[n] = {'system_type': section, 'init': init, 'params': params, 'tspan': tspan, 'solver_args': solver_args}
        
        n += 1
        
    return config_data

###############################################################################
#Main Set Up
###############################################################################    
def get_models_and_params(filename, user_core_models={}):
    '''Reads the config file and adds combines it with core_model data. If you are
    using a core model that is not in the database, you must provide the core_model
    using the core_models argument where the key is the system_type. Returns the 
    models, params and config_data.
    
    :param filename: Name of file to read.
    :type filename: str
    :param user_core_model: A dictionary of core_models indexed by their system_type.
        core_models already in the database do not need to be specified here.
    :type user_core_model: dict, optional
    '''

    config_data, core_models = setup_helper(filename, from_config, user_core_models)
    models, params           = compile_models(core_models, config_data)
    return models, params, config_data

def compile_models(core_models, config_data):
    models     = make_compiled_models_template(core_models)
    params     = {}
    
    for key in config_data:
        models[key]['init']                    = config_data[key].get('init',       {1: np.array([0]*len(models[key]['states'])) })
        models[key]['tspan']                   = config_data[key].get('tspan',      [np.linspace(0, 600, 31)])
        models[key]['int_args']['solver_args'] = config_data[key].get('solver_args', {})
        
        try:
            temp       = {param + '_' + str(key): config_data[key]['params'][param].values  for param in config_data[key]['params']}
        except:
            temp       = {param + '_' + str(key): config_data[key]['guess'][param].values  for param in config_data[key]['guess']}
            
        params = {**params, **temp}
    
    return models, pd.DataFrame(params)

def setup_helper(filename, reader, user_core_models={}):
    '''
    The first return value is the config_data which will be generated based on
    three possible scenarios:
    
        1. filename is a dictionary 
           The filename is already config_data. No processing needed.
           
        2. filename is a string
           filename is name of the settings file to be opened and parsed by reader. 
           
        3. filename is neither of the above
           filename is an iterable containing names of settings file. It will be
           iteratively parsed by reader and subsequently reindexed to give config_data.
    
    The second return value is a list of core_model data structures associated with 
    config_data. The list is arranged in the order given by config_data and the 
    core_models are retrieved based on two possible scenarios:
        
        1. The system_type in config_data[model_num]['system_type'] is in core_models
           The core_model indexed under core_models[system_type] will be used.
           
        2. The system_type in config_data[model_num]['system_type'] is not in core_models
           The function searches the BMSS database for the appropriate core_model
           using quick_search
    
    :meta: private
    '''
    if type(filename) == str:
        config_data = reader(filename)
    elif type(filename) == dict:#Assume user has already imported the data
        config_data = filename
    else:#Assume data is spread across multiple files
        config_data = [value for f in filename for value in reader(f).values()]
        config_data = dict(enumerate(config_data, start=1))
    
    core_models = [user_core_models[config_data[key]['system_type']] if config_data[key]['system_type'] in user_core_models else mh.quick_search(config_data[key]['system_type']) for key in config_data]
    core_models = [mh.copy_core_model(core_model) for core_model in core_models]
    
    return config_data, core_models

def make_compiled_models_template(core_models):
    '''Constructs a compiled models dict using a list/tuple of core models
    
    :meta private:
    '''
    models = {}
    
    for i in range(len(core_models)):
        core_model     = core_models[i]
        model_function = mh.get_model_function(core_model['system_type'], local=False)
        
        temp = {'function' : model_function,
                'init'     : {},
                'states'   : core_model['states'],
                'params'   : [param+'_'+str(i+1) for param in core_model['parameters'] + core_model['inputs']],
                'tspan'    : [],
                'int_args' : {'modify_init'   : None,
                              'modify_params' : None,
                              'solver_args'   : {},
                              }
                }
        
        models[i+1] = temp
    
    return models

###############################################################################
#Template Generation
###############################################################################
def make_settings_template(system_types_settings_names, filename='', user_core_models={}):
    '''Writes settings to a config file. If you are using a core_model that is 
    not in the database, you must provide the core_model using the core_models 
    argument where the key is the system_type.Returns the code as a string.
    
    :param system_types_settings_names: Pairs of tuples containing (system_type, settings_name)
    :type system_types_settings_names: list or tuple
    :param filename: The name of the file to write to
    :type filename: str, optional
    :param user_core_model: A dictionary of core_models indexed by their system_type.
        core_models already in the database do not need to be specified here.
    :type user_core_model: dict, optional
    '''
    result = ''
    
    system_types_settings_names1 = [system_types_settings_names] if type(system_types_settings_names) == str else system_types_settings_names
    
    for pair in system_types_settings_names1:
        try:
            system_type, settings_name = pair
        except:
            system_type, settings_name = pair, '__default__'

        core_model     = user_core_models[system_type] if system_type in user_core_models else mh.quick_search(system_type)
        parameters     = core_model['parameters'] + core_model['inputs']
        states         = core_model['states']
        section_header = '[' + core_model['system_type'] + ']'
        
        settings = {'system_type'     : system_type, 
                    'settings_name'   : settings_name,
                    'parameters'      : {},
                    'units'           : {},
                    'init'            : {},  
                    'priors'          : {}, 
                    'parameter_bounds': {},
                    'tspan'            : [],
                    'fixed_parameters' : [],
                    'solver_args'      : {'rtol'   : 1.49012e-8,
                                          'atol'   : 1.49012e-8,
                                          'tcrit'  : [],
                                          'h0'     : 0.0,
                                          'hmax'   : 0.0,
                                          'hmin'   : 0.0,
                                          'mxstep' : 0
                                          },
                    }
        
        if settings_name:
            settings = sh.quick_search(system_type, settings_name, skip_constructor=True)
        
        longest       = len(max(parameters,               key=len))    
        longest_      = len(max(settings['solver_args'],  key=len))  
        solver_args   = settings['solver_args'].keys()
        init_values   = dict_template('init',             states,     longest, settings['init'])
        param_values  = dict_template('parameter_values', parameters, longest, settings['parameters'])
        units_values  = dict_template('units',            parameters, longest, settings['units'], default='')
        tspan         = ['[' + ', '.join(['{:.2f}'.format(x) for x in segment]) + ']' for segment in settings['tspan']]
        tspan         = t_template('tspan',            '[' + ', '.join(tspan) + ']')
        solver_args   = dict_template('solver_args',      solver_args, longest_, settings['solver_args'])
        
        model_id         = '#id = ' + str(core_model['id'])
        model_equations  = '#equations = \n' + '\n'.join(['#\t' + line for line in core_model['equations'] ])
        section_header   = '\n'.join([section_header, model_id, model_equations])
        
        result += '\n\n'.join([section_header, init_values, param_values, units_values, tspan, solver_args, '']) 
        
    if filename:
        with open(filename, 'w') as file:
            file.write(result)
    return result

def dict_template(sub_section, keys, longest, data={}, default='[]'):
    '''
    :meta private:
    '''
    result  = sub_section + ' = \n' + ',\n'.join(['\t'  + key + ' '*(longest - len(key)) + ' = ' + str(data.get(key, default)) for key in keys])
    
    return result

def t_template(sub_section, values):
    '''
    :meta private:
    '''
    
    result = sub_section + ' = \n' + '\n\t' + values
    return result
    
def list_template(sub_section, values):
    '''
    :meta private:
    '''

    result = sub_section + ' = \n' + '\n\t' + '[' +', '.join([str(value) for value in values]) + ']'

    return result
                       
if __name__ == '__main__':
    pass


import keyword
import string
import warnings
from   pathlib import Path
from   runpy   import run_path

###############################################################################
#Non-Standard Imports
###############################################################################
try:
    from .model_coder import equations_to_code, model_to_code
except:
    from model_coder import equations_to_code, model_to_code
    
punctuation = string.punctuation.replace('_', '').replace('.', '')

def check_model_terms(model):
    states    = model['states']
    params    = model['parameters']
    inputs    = model.get('inputs', [])
    equations = model['equations']
    states_set = set(states)
    params_set = set(params)
    inputs_set = set(inputs)
    if len(states_set) != len(states):
       return False, 'Multiple occurence of certain states in states.'
    if len(params_set) != len(params):
        return False, 'Multiple occurence of certain parameters in parmeters.'
    if len(inputs_set) != len(inputs):
        return False, 'Multiple occurence of certain inputs in inputs.'
    
    shared = states_set.intersection(params_set, inputs_set)
    if shared:
        return False, 'Found ' + str(shared) + ' in both states, parameters and/or inputs.'
    
    terms    = list(map(get_terms, equations))
    lhs, rhs = [set().union(*s) for s in zip(*terms)]
    
    check_illegal_terms(terms)
    
    defined = lhs.union(states_set, params_set, inputs_set)
    defined.update(['ln', 'log', 'np', 'array', 't'])
    
    def defined_test(term):
        if term in defined or 'model_' in term or 'np.' in term[:3]:
            return True
        return False
    
    undefined = [term for term in rhs if not defined_test(term)] 
    
    if undefined:
        return False, 'Undefined terms ' + str(undefined)
    
    unused = [term for term in inputs+params if term not in rhs and '_init' not in term]
    
    if unused:
        return False, 'Unused terms ' + str(unused)
    
    else:
        temp    = states+params+inputs
        var     = ','.join(temp) + '= np.random.rand(' + str(len(temp)) + ')*10'
        conv    = ','.join(temp) + '= list(map(float, [' + ','.join(temp) + ']))'
        y_test  = 'y = [' + ','.join(states) + ']'
        t_test  = 't = 0\ndt = 1e-3'
        p_test  = 'params = [' + ','.join(params+inputs) + ']'
        call    = 'y = y + dt*model_' + model['system_type'].replace(', ', '_') + '(y, t, params)'
        
        try:
            eq     = model_to_code(model, use_numba=False)
        except Exception as e:
            return False, str(e.args[0])
        
        test_string = '\n\n'.join([eq, var, conv, y_test, t_test, p_test, call, call])
        
        filename    = Path(__file__).parent/ 'checker_do_not_run.py'
        for i in range(10):
            try:
                
                with open(filename, 'w') as file:
                    file.write(test_string)
                run_path(filename)
                
            except ZeroDivisionError:
                if i < 9:
                    pass
                else:
                    return False, 'Division by zero detected.'
            except IndexError:
                return False, 'Length of return value is not equal to length of states.'
            except NameError as e:
                name = e.args[0].split(' ')[1]
                if 'model_' in name[:7]:
                    system_type = model["system_type"]
                    warnings.warn(f'{system_type} may require one or more submodels.')
                else:
                    return False, str(e.args[0])
            except Exception as e:
                return False, str(e.args[0])
            else:
                break
            
    return True, ''

def get_terms(f):
    '''
    Supporting function for make_group. Extracts all terms in an expression.
    '''
    global punctuation
    
    fs = f.strip()
    fs = f.split('#')[0]
    
    if not fs:
        return set(), set()
    
    eq = [s.strip() for s in  fs.split('=', 1)]
    
    if len(eq) < 2:
        raise Exception('Line is not an equation: ' + f)

    if any(list(map(illegal_equal, eq))):
        raise Exception('Syntax error. Illegal use of "=" symbol in line: '+ f)
    
    tr = str.maketrans(punctuation, ','*len(punctuation))
    eq = [set([i for i in s.replace(' ', '').translate(tr).split(',') if i and not isnum(i)]) for s in eq]
    
    return eq

def isnum(x):
    try:
        float(x)
        return True
    except:
        pass
    try:
        float(x+'0')
        return True
    except:
        return False

def illegal_equal(x):
    S = x.split('==')
    if S[0] == '' or S[-1]=='':
        return True
    
    for i in range(len(S)):
        s = S[i].split('=')

        if len(s) > 1:
            return True
    return False

def check_illegal_terms(terms):
    for term in terms:
        for illegal_term in ['np']:
            if term == illegal_term:
                raise Exception('Illegal term detected: ' + term)
        
        for illegal_term in keyword.kwlist:
            if term == illegal_term:
                raise Exception('Illegal term detected: ' + term)
                
    return
    
if __name__ == '__main__':
    __model__ = {'id'          : 'bmss01001',
                 'system_type' : 'Inducible, ConstInd',
                 'states'      : ['mRNA', 'Pep'], 
                 'parameters'  : ['syn_mRNA', 'deg_mRNA', 'syn_Pep', 'deg_Pep', 'Ki'],
                 'inputs'      : ['Ind'],
                 'equations'   : ['dmRNA = syn_mRNA*Ind/(Ind + Ki) - deg_mRNA*mRNA',
                                  '',
                                  'dPep  = syn_Pep*mRNA - 1/(deg_Pep)'
                                  ],
                 'ia'          : ''
                 
                 }
    
    __model__ = {'id'          : 'bmss01002',
                  'system_type' : 'BMSS, GrowthModel, MortalityPhase',
                  'states'      : ['OD'], 
                  'parameters'  : ['N0', 'm1', 'm2', 'tc1', 'tc2'],
                  'inputs'      : ['t'],
                  'equations'   : ['a = 1 / (tc1 ** m1)',
                                  'b = 1 / (tc2 ** m2)',
                                  '',
                                  'dOD = N0 * (a * m1 * (t ** (m1-1)) - b * m2 * (t ** (m2-1))) * (2.7183**(a * (t ** m1) - b * (t ** m2)))',
                                  ],
                  'ia'          : ''
                 
                  }
    
    r = check_model_terms(__model__)
    print(r)


import configparser
import numpy    as np
import os       as os
import os.path  as osp
from   pandas   import concat, DataFrame, Series, read_sql_query

###############################################################################
#Non-Standard Import
###############################################################################
try:
    from . import model_handler as mh
    from . import settings_checker as sc
except:
    import model_handler as mh
    import settings_checker as sc
    
###############################################################################
#Globals
###############################################################################
_dir = osp.dirname(osp.abspath(__file__))

table_sql = '''
CREATE TABLE IF NOT EXISTS "settings" (
	"system_type"      TEXT,
    "settings_name"    TEXT,
    "units"            TEXT,
    "parameters"       TEXT,
    "init"             TEXT,
    "priors"           TEXT,
    "parameter_bounds" TEXT,
    "fixed_parameters" TEXT, 
    "tspan"            TEXT,
    "solver_args"      TEXT,
    "active"           INTEGER DEFAULT 1,
    UNIQUE(system_type, settings_name)
);
'''
###############################################################################
#Parameter Ensemble Constructor
###############################################################################    
def make_settings(system_type,    settings_name,       units,    
                  parameters,     init=None,           tspan=None,           
                  priors={},      parameter_bounds={}, fixed_parameters=[], 
                  solver_args={}, init_orient='scenario', 
                  user_core_model=None, 
                  **ignored):
    '''Checks and standardizes formatting of data types. Ignores irrelavant keyword arguments.
    Returns a settings data structure which is a dict.
    
    :param system_type: A string of keywords serving as a unique identifier for 
        the core_model separated by commas, will be formatted so there is one space
        after each comma, keywords should be in CamelCase
    :type system_type: str
    :param settings_name: A string, forms a unique identifier when paired with system_type
    :type settings_name: str
    :param units: A dict where the keys are parameter names and the values are the units in strings
    :type units: dict
    :param parameters: A pandas DataFrame of parameter values OR a dictionary, 
        column names/keys must match the parameter names of the corresponding core_model  
    :type parameters: pandas DataFrame or dict
    :param init: A dictionary where the keys correspond to a state in the core_model and 
        the values are the initial values for numerical integration OR where the keys
        correspond to different scenarios 
    :type init: dict
    :param tspan: A list of segments for piecewise numerical integration where each segments
        is a list of time points, defaults to values generated using numpy.linspace
    :type tspan: list, optional
    :param priors: A dictionary of Gaussian priors for parameter estimation where each key 
        is a parameter name and each value a tuple in the form (mean, standard_deviation)
    :type priors: dict, optional
    :param parameter_bounds: A dictionary of parameter bounds for parameter estimation 
        where each key is a parameter name and each value a tuple in the form (lower, upper)
    :type parameter_bounds: dict, optional
    :param fixed_parameters: A list of parameter names that will be fixed during parameter estimation.
    :type fixed_parameters: list, optional 
    :param solver_args: Additional keyword arguments for scipy.integrate.odeint, defaults to None
    :type solver_args: dict, optional
    :param init_orient: 'scenario' if the keys in init are scenario numbers, 'state' 
        if the keys are the state names, defaults to 'scenario'
    :type init_orient: str
    :param user_core_model: A core_model data structure for cross-checking. If None, 
        this function will search the database for the core_model.
    :type user_core_model: dict
    '''
    
    system_type1 = system_type if type(system_type) == str else ', '.join(system_type)
    if user_core_model:
        core_model   = user_core_model
        if core_model['system_type'] != system_type1:
            raise Exception('system_type does not match that of user_core_model.')
        print('Making settings using user_core_model')
    else:
        core_model = mh.quick_search(system_type1)
        
    param_values = sc.check_and_assign_param_values(core_model, parameters)
    init1        = sc.check_and_assign_init(core_model['states'], init, init_orient)
    tspan1       = sc.check_and_assign_tspan(tspan)
    priors1      = sc.check_and_assign_priors(param_values, priors)
    bounds1      = sc.check_and_assign_parameter_bounds(param_values, parameter_bounds)
    solver_args1 = sc.check_and_assign_solver_args(solver_args)

    if fixed_parameters:
        fixable_parameters = core_model['parameters'] + core_model['inputs']
        if not all([p in fixable_parameters for p in fixed_parameters]):
            raise Exception('Error in fixed parameters. Unexpected parameters found.')
    
    
    settings_name1 = '__default__' if settings_name == '_' else settings_name
    
    result       = {'system_type'      : core_model['system_type'],
                    'settings_name'    : settings_name1,
                    'units'            : units,
                    'parameters'       : param_values,
                    'init'             : init1,
                    'priors'           : priors1,
                    'parameter_bounds' : bounds1,
                    'tspan'            : tspan1,
                    'fixed_parameters' : fixed_parameters,
                    'solver_args'      : solver_args1,
                    }
    return result

###############################################################################
#Param Storage
###############################################################################
def add_to_database(settings, dialog=False):
    '''Accepts a settings data structure and adds it to UBase. dialog is currently ignored.
    '''
    
    if dialog:
        raise NotImplemented('Dialog option has not been implemented for this function.')
    
    if (settings['system_type'], settings['settings_name']) in list_settings(database=mh.MBase):
        raise Exception('Settings exists in MBase and cannot be modified. Please use a different settings_name.')

    return backend_add_to_database(settings, database=mh.UBase, dialog=dialog)

###############################################################################
#Supporting Functions
###############################################################################
def backend_add_to_database(settings, database, dialog=False):
    '''Supporting function for add_to_database. Do not run.
    
    :meta private:
    '''
    
    #Check core model exists
    core_model = mh.quick_search(settings['system_type'])
    
    #Prevent duplicates between MBase and UBase
    system_type_ensemble_name = (settings['system_type'], settings['settings_name'])
    
    if database == mh.MBase and system_type_ensemble_name in list_settings(database=mh.UBase):
        raise Exception(str(system_type_ensemble_name) + ' cannot be added to MBase as it is already in UBase.')
        
    if database == mh.UBase and system_type_ensemble_name in list_settings(database=mh.MBase):    
        raise Exception(str(system_type_ensemble_name) + ' cannot be added to UBase as it is already in MBase.')
    
    params_dict = settings['parameters'].to_dict('list')
    params_dict = {key: params_dict[key] for key in settings['parameters'].to_dict('list')}
    
    row = {'system_type'   : settings['system_type'],
           'settings_name' : settings['settings_name'],
           'units'         : str(settings['units']),
           'parameters'    : str(params_dict),
           }
    
    for key1 in ['init', 'priors', 'parameter_bounds']:
        row[key1] = str({key2: list(settings[key1][key2]) for key2 in settings[key1]})
    
    row['tspan']            = str([list(span) for span in settings['tspan']])
    row['fixed_parameters'] = str(settings['fixed_parameters'])
    row['solver_args']      = str(settings['solver_args'])
    
    mh.add_row('settings', row, database)
    print('Added settings ' + row['settings_name'])
    return row['system_type'], row['settings_name']

###############################################################################
#Search
###############################################################################    
def search_database(system_type='', settings_name='', database=None, skip_constructor=False, active_only=True):
    '''Searches database for settings data structure based on system_type and settings
    name. Returns a list of settings dictionaries.
    
    :param system_type: A string that will be matched against entries in the database
    :type system_type: str
    :param settings_name: A string corresponding to any key in the core_model, will 
        be used for matching
    :type settings_name: str
    :param database: Can be either MBase or UBase, if None, this function will search
        both databases, defaults to None
    :type database: SQL Connection, optional
    :param skip_constructor: For backend use, defaults to False
    :type skip_constructor: bool, optional
    :param active_only: A boolean for backend use, if True, this limits the search 
        to rows where the value of active is True, defaults to True
    :type active_only: bool, optional
    '''
    #Can only add a settings for active models
    core_model = mh.quick_search(system_type, active_only=True)
    result     = []
    
    if settings_name and system_type:
        comm = 'SELECT * FROM settings WHERE system_type LIKE "%' + system_type + '%" AND settings_name LIKE "%' + settings_name
    elif settings_name:
        comm = 'SELECT * FROM settings WHERE settings_name LIKE "%' + settings_name
    else:
        comm = 'SELECT * FROM settings WHERE system_type LIKE "%' + system_type  
    
    if active_only:
        comm += '%" AND active = 1;'
    else:
        comm += '%";'
        
    databases = [database] if database else [mh.MBase, mh.UBase]
    
    for database in databases:
        with database as db:
            cursor       = db.execute(comm)
            all_settings = cursor.fetchall()    
         
        columns = database.execute('PRAGMA table_info(settings);')
        columns = columns.fetchall()
        columns = [column[1] for column in columns if column[1]!='active']
        all_settings  = [dict(zip(columns, s[:-1])) for s in all_settings]
        
        for s in all_settings:
            s['units']            = eval(s['units'])
            s['parameters']       = eval(s['parameters'])
            s['init']             = eval(s['init'])
            s['priors']           = eval(s['priors'])
            s['parameter_bounds'] = eval(s['parameter_bounds'])
            s['tspan']            = eval(s['tspan'])
            s['fixed_parameters'] = eval(s['fixed_parameters'])
            s['solver_args']      = eval(s['solver_args'])
        
        if skip_constructor:
            result += all_settings
        else:
            result += [make_settings(**s) for s in all_settings]

    return result

def quick_search(system_type='', settings_name='', error_if_no_result=True, **kwargs):
    '''Searches database and returns first search result. If no settings name is 
    given, will attempt to search for the settings indexed under "__default__".
    '''
    
    try:
        settings_name1 = settings_name if settings_name else '__default__'
        settings = search_database(system_type, settings_name1, **kwargs)[0]
    except:
        if error_if_no_result:
            raise Exception('Could not find default parameters for system_type '  + str(system_type))
        else:
            return
        
    if settings['system_type'] != system_type or settings['settings_name'] != settings_name:
        if error_if_no_result:
            raise Exception('Could not find default parameters for system_type '  + str(system_type))
        else:
            return
        
    return settings
    
def list_settings(database=None):
    '''Returns a list of tuples in the form (system_type, settings_name) from 
    database.
    '''
    if database:
        comm         = 'SELECT system_type, settings_name FROM settings where active = 1'
        cursor       = database.execute(comm)
        all_settings = [s for s in cursor.fetchall()]
        return all_settings
    else:
        return list_settings(mh.MBase) + list_settings(mh.UBase)
    

###############################################################################
#Interfacing with Pandas
###############################################################################
def to_df(database=None):
    '''Returns a copy of the databases as a pandas DataFrame.
    
    :param database: The database to be read, if None, both databases will be read,
        defaults to None
    :type database: SQL Connection
    '''
    if database:
        df = read_sql_query("SELECT * from settings", database)
        return df
    else:

        df = to_df(mh.MBase), to_df(mh.UBase)
        df = concat(df, ignore_index=True)
        return df

def from_df_replace(df, database):
    '''
    For backend maintenance only. Do not run unless you know what you are doing.
    
    :meta private:
    '''
    return df.to_sql('settings', database, if_exists='replace', index=False)

###############################################################################
#Interfacing with Configparser
###############################################################################    
def from_config(filename, user_core_models={}):
    '''Returns a settings data structure.
    
    :param filename: Name of file to read.
    :type filename: str
    :param user_core_models: A dictionary of core_models required for calling make_settings,
        to be ignored if the core_model is already in the database.
    :type user_core_models: dict, optional
    '''
    
    config              = configparser.RawConfigParser()
    config.optionxform  = lambda option: option
    all_settings        = []
    with open(filename, 'r') as file:
        config.read_file(file)
    
    for section in config.sections():
        if section in ['id', 'system_type', 'states', 'parameters', 'inputs', 'equations', 'ia']:
            continue
        if 'system_type' not in config[section]:
            continue
        
        system_type   = config[section]['system_type']
        params        = config[section]['parameter_values']
        units         = config[section]['units']
        init          = config[section].get('init', {})
        priors        = config[section].get('priors', {})
        bounds        = config[section].get('parameter_bounds', {})
        tspan         = config[section].get('tspan', '[[0, 600, 31]]')
        fixed_parameters = config[section].get('fixed_parameters', '[]')
        settings_name = str(section)
        
        core_model    = mh.quick_search(system_type)
        system_type   = core_model['system_type']
        states        = core_model['states']
        if not settings_name:
            raise Exception('Ensemble name not valid!')
        settings_name = '__default__' if settings_name == '_' else settings_name
        
        if init:
            if '=' in init:
                init = string_to_dict(init)
                try: 
                    init        = DataFrame(init)
                    init.index += 1
                    init        = init.T.to_dict('list')
                except:
                    init  = {1: list(init.values())}

                for key in init:
                    if len(init[key]) != len(states):
                        raise Exception('Length of init must match number of states. system_type given: ' + str(system_type))
            else:
                init = eval(init)
                if len(init) != len(states):
                    raise Exception('Length of init must match number of states. system_type given: ' + str(system_type))
                init = {1: init}

        params           = string_to_dict(params)
        units            = string_to_dict(units)
        priors           = string_to_dict(priors) if priors else {}
        bounds           = string_to_dict(bounds) if bounds else {}
        fixed_parameters = string_to_list_string(fixed_parameters)
        tspan            = eval_tspan_string(tspan)
        
        settings = {'system_type'      : system_type, 
                    'settings_name'    : settings_name,
                    'parameters'       : params,
                    'units'            : units,
                    'init'             : init,  
                    'priors'           : priors, 
                    'parameter_bounds' : bounds,
                    'tspan'            : tspan,
                    'fixed_parameters' : fixed_parameters
                    }
        
        user_core_model = user_core_models.get(settings['system_type'], {})
        all_settings.append(make_settings(**settings, user_core_model=user_core_model))
        
    return all_settings

def eval_tspan_string(string):
    try:
        return string_to_linspace(string)
    except:
        return np.array(eval(string.replace('\n', '')))

def string_to_linspace(string):
    try:
        return [np.linspace(*segment) for segment in eval('[' + string +']')]
    except:
        return [np.linspace(*segment) for segment in eval(string)]

def string_to_list_string(string):
    temp = string.strip().replace('[', '').replace(']', '')
    temp = [s.strip() for s in temp.split(',')]
    
    return [s for s in temp if s]
    
def string_to_dict(string):
    result = [s.strip() for s in split_at_top_level(string)]
    result = [line.split('=') for line in result ]
    result = [[lst[0].strip(), '='.join(lst[1:]).strip()] for lst in result]
    
    result = {pair[0]: try_eval(pair[1]) for pair in result}    
    
    return result

def split_at_top_level(string, delimiter=','):
    nested = []
    buffer = ''
    result = []
    matching_bracket = {'(':')', '[':']', '{':'}'}
    
    for char in string:
        if char in ['[', '(', '{']:
            nested.append(char)
            buffer += char
        
        elif char in [']', ')', '}']:
            if char == matching_bracket.get(nested[-1]):
                nested = nested[:-1]
                buffer += char
            else:
                raise Exception('Mismatched brackets.' )
        elif char == delimiter and not nested:
            if buffer:
                result.append(buffer)
                buffer = ''
        else:
            buffer += char
    if buffer:
        result.append(buffer)
    return result

def try_eval(x):
    try:
        return eval(x)
    except:
        return x
    
###############################################################################
#Template Generation
###############################################################################
def make_settings_template(system_types_settings_names, filename=''):
    '''Writes settings to a config file. Returns the code as a string.
    
    :param system_types_settings_names: Pairs of tuples containing (system_type, settings_name)
    :type system_types_settings_names: list or tuple
    :param filename: The name of the file to write to
    :type filename: str, optional
    '''
    result = ''
    
    system_types_settings_names1 = [system_types_settings_names] if type(system_types_settings_names) == str else system_types_settings_names
    
    for pair in system_types_settings_names1:
        try:
            system_type, settings_name = pair
        except:
            system_type, settings_name = pair, '__default__'
        
        if type(system_type) == str:
            core_model     = mh.quick_search(system_type)
        else:
            core_model  = system_type
            system_type = core_model['system_type']
            
        parameters     = core_model['parameters'] + core_model['inputs']
        states         = core_model['states']
        section_header = '[]\nsystem_type = ' + core_model['system_type']
        
        if settings_name:
            settings = quick_search(system_type, settings_name, skip_constructor=True)

        else:            
            settings = {'system_type'      : system_type, 
                        'settings_name'    : settings_name,
                        'parameters'       : {},
                        'units'            : {},
                        'init'             : {},  
                        'priors'           : {}, 
                        'parameter_bounds' : {},
                        'tspan'            : [],
                        'fixed_parameters' : [],
                        'solver_args'      : {'rtol'   : 1.49012e-8,
                                              'atol'   : 1.49012e-8,
                                              'tcrit'  : [],
                                              'h0'     : 0.0,
                                              'hmax'   : 0.0,
                                              'hmin'   : 0.0,
                                              'mxstep' : 0
                                              },
                        }
        
        settings['init'] = DataFrame.from_dict(settings['init'], orient='index', columns=states).to_dict('list')
        longest          = len(max(parameters,               key=len))
        longest_         = len(max(settings['solver_args'],  key=len))  
        solver_args      = settings['solver_args'].keys()
        init_values      = dict_template('init',             states,     longest, settings['init'])
        param_values     = dict_template('parameter_values', parameters, longest, settings['parameters'])
        prior_values     = dict_template('priors',           parameters, longest, settings['priors'])
        bounds_values    = dict_template('parameter_bounds', parameters, longest, settings['parameter_bounds'])
        units_values     = dict_template('units',            parameters, longest, settings['units'], default='')
        tspan            = list_template('tspan',            [list(segment) for segment in settings['tspan']])
        fixed_parameters = list_template('fixed_parameters', settings['fixed_parameters'])
        solver_args      = dict_template('solver_args',      solver_args, longest_, settings['solver_args'])
        model_id         = '#id = ' + str(core_model['id'])
        model_equations  = '#equations = \n' + '\n'.join(['#\t' + line for line in core_model['equations'] ])
        section_header   = '\n'.join([section_header, model_id, model_equations])
        
        result += '\n\n'.join([section_header, init_values, param_values, prior_values, bounds_values, units_values, tspan, fixed_parameters, solver_args])
        
    if filename:
        with open(filename, 'w') as file:
            file.write(result)
    return result

def dict_template(sub_section, keys, longest, data={}, default='[]'):
    result  = sub_section + ' = \n' + ',\n'.join(['\t'  + key + ' '*(longest - len(key)) + ' = ' + str(data.get(key, default)) for key in keys])
    
    return result

def list_template(sub_section, values):
    return sub_section + ' = \n' + '\n\t'+ str(values)

###############################################################################
#Direct Config to Database
###############################################################################
def config_to_database(filename):
    '''Adds settings in file to the database. Returns a list of tuples of 
    system_type, settings_name pairs.
    '''

    all_settings = from_config(filename)
    
    result = list(map(add_to_database, all_settings))
    
    return result

def backend_config_to_database(filename, database):
    '''
    For backend maintenance. Do not run.
    
    :meta private:
    '''
    all_settings = from_config(filename)
    
    
    result = [backend_add_to_database(settings, database) for settings in all_settings]
    
    return result

###############################################################################
#Modification
###############################################################################
def _backend_modify_database(system_type, settings_name, data):
    '''
    Updates the row containing system_type with data.
    Automatically determines if the system_type is part of MBase or UBase.
    Not meant to be used on unsanitized data.
    '''
            
    if 'BMSS' in system_type:
        database = mh.MBase
    else:
        database = mh.UBase
    
    with database as db:
        for column, value in data.items():
            comm = "UPDATE settings SET " + str(column) + " = '" + str(value) + "' WHERE system_type = '" + str(system_type) + "' AND settings_name = '" + str(settings_name) + "'"
            cur  = db.cursor()
            cur.execute(comm)
            print('Changes to ' + str(system_type) + ' saved to database.')
        
    return

###############################################################################
#Deletion
###############################################################################
def delete(system_type, settings_name):
    '''Sets the value of the active column of the core_model in the .db file to 0.
    Can be restored using the restore function.
    '''
    database = mh.UBase
    with database as db:
        comm = 'UPDATE settings SET active = 0 WHERE system_type = "' + system_type + '" AND settings_name = "' + settings_name + '";'
        cur  = db.cursor()
        cur.execute(comm)
        
def restore(system_type, settings_name):
    '''Sets the value of the active column of the core_model in the .db file to 1.
    '''
    database = mh.UBase
    with database as db:
        comm = 'UPDATE settings SET active = 1 WHERE system_type = "' + system_type + '" AND settings_name = "' + settings_name + '";'
        cur  = db.cursor()
        cur.execute(comm)
    
def true_delete(system_type, settings_name, database):
    if not quick_search(system_type, settings_name, active_only=False, error_if_no_result=False, database=database):
        print(str(system_type) + '/' + str(settings_name) + ' could not be deleted as it was not found.')
    
    with database as db:
        comm = 'DELETE FROM settings WHERE system_type="'+ system_type + '" AND settings_name="' + settings_name +'";'    
        db.execute(comm)
        print('Removed ' + str(system_type) + '/' + str(settings_name))
    
###############################################################################
#Function for Setup
###############################################################################
def setup():
    for database in [mh.MBase, mh.UBase]:
        res         = database.execute("SELECT name FROM sqlite_master WHERE type='table';")
        table_names = [table[0] for table in res]
        if 'settings' not in table_names:
            mh.create_table(database, table_sql)
    
    print('Connected to MBase_settings, UBase_settings')
    return database

###############################################################################
#Initialization
###############################################################################
setup()

if __name__ == '__main__':
    from pathlib import Path
    
    filename = Path(os.getcwd()) / 'BMSS_markup' / 'Monod_Inducible_Bioconversion_ProductInhibitedGrowth.ini'
    
    core_model = mh.from_config(filename)
    settings   = from_config(filename)
    
    
    
    __model__ = {'system_type' : ['DUMMY', 'DUMMY'],
                 'states'      : ['m', 'p'], 
                 'parameters'  : ['synm', 'degm', 'synp', 'degp'],
                 'inputs'      : ['Ind'],
                 'equations'   : ['dm = synm*Ind - degm*m',
                                  'dp  = synp*m - degp'
                                  ],
                 'ia'          : 'ia_result_bmss01001.csv'
                 }
    
    __settings__ = {'system_type'      : ['DUMMY, DUMMY'],
                    'settings_name'    : '__default__',
                    'units'            : {'synm' : 'M/min', 
                                          'degm' : '1/min',
                                          'synp' : '1/min',
                                          'degp' : '1/min',
                                          'Ind'  : 'M'
                                          },
                    'parameters'       : {'synm' : np.array([0.02]), 
                                          'degm' : np.array([0.15]),
                                          'synp' : np.array([0.02]),
                                          'degp' : np.array([0.012]) ,
                                          'Ind'  : np.array([1.0]) 
                                          },
                    'init'             : {1: [0, 0]},
                    'parameter_bounds' : {'degm' : [0.001, 0.03], 
                                          'degp' : [0.01, 0.5], 
                                          },
                    'solver_args'      : {'rtol'   : 1.49012e-8,
                                          'atol'   : 1.49012e-8,
                                          'tcrit'  : [],
                                          'h0'     : 0.0,
                                          'hmax'   : 0.0,
                                          'hmin'   : 0.0,
                                          'mxstep' : 0
                                          },
                }
    
    # core_model = mh.make_core_model(**__model__)
    # settings   = make_settings(**__settings__, user_core_model=core_model)
    
    ###############################################################################
    #Testing Parameter Formats
    ###############################################################################
    # __settings__['parameters'] = {key+'_1' : __settings__['parameters'][key] for key in __settings__['parameters']}
    # settings = make_settings(**__settings__, user_core_model=core_model)
    
    # __settings__['parameters'] = Series(__settings__['parameters'], name=500)
    # settings = make_settings(**__settings__, user_core_model=core_model)
    
    ###############################################################################
    #Testing Initial Value Formats
    ###############################################################################
    # __settings__['init'] = {'m' : [0],
    #                         'p' : [0],
    #                         }
    # settings = make_settings(**__settings__, init_orient='states', user_core_model=core_model)
    
    # __settings__['init'] = [[0, 0], [0, 0]]
    # settings = make_settings(**__settings__, init_orient='states', user_core_model=core_model)
    
    # __settings__['init'] = Series([0, 0], index=['m', 'p'])
    # settings = make_settings(**__settings__, init_orient='states', user_core_model=core_model)
    
    # __settings__['init'] = DataFrame([[0, 0]], columns=['m', 'p'], index=[1])
    # settings = make_settings(**__settings__, init_orient='states', user_core_model=core_model)
    
    ###############################################################################
    #Testing Parameter Bound Formats
    ###############################################################################
    # __settings__['parameter_bounds'] = {}
    # settings = make_settings(**__settings__, user_core_model=core_model)
    
    
    # add_to_database(settings)
    
    



# -*- coding: utf-8 -*-
"""
This module contains functions to generate COMBINE OMEX files.

Features:
-to generate COMBINE OMEX file from a BMSS database
"""

import tellurium as te, os
from pathlib import Path
import glob

import BMSS.sbml.simplesbml as simplesbml
import BMSS.sbml.sbmlgen as sbmlgen
import BMSS.models.model_handler    as mh
import BMSS.models.settings_handler as sh
import phrasedml


#--- Main Body ---
def database_to_combine(system_type, settings_name, Plot_Variable, outputpath='', KISAO_algorithm='0'):
    '''Select model stored in database and generate the SBML file. 
    
    :param system_type: system type of model in string format
    :param settings_name: settings name of model in string format
    :param Plot_Variable: Variables to be plotted in list format
    :param outputpath:  filepath to output OMEX file in OS path format
    :param KISAO_algorithm:  Kinetic Simulation Algorithm Ontology number in string format   
    '''
    
    combinefilelist=[]
    number_scenario = 0

    
    search_result_settings  = sh.search_database(system_type, settings_name)
    core_model              = mh.quick_search(system_type)
    settings                = search_result_settings[0]
    
    #-- Checkers --
    plotvariablechecker(Plot_Variable, core_model)
    KISAOchecker(KISAO_algorithm)
    if system_type not in mh.list_models():
        raise Exception("Model is not in database")
    
    combinefilename, number_scenario = Combinecreator(core_model, settings, Plot_Variable, outputpath, KISAO_algorithm)
    combinefilelist.append(combinefilename)
   
    
        
    print()
    print('Number of Scenarios in COMBINE File=', number_scenario)
    print('COMBINE files outputed: ', combinefilelist)    
    
    return combinefilelist
    
#--- COMBINE Generator and Exporter ---
def Combinecreator(core_model, settings, Plot_Variable, outputpath='', KISAO_algorithm='0'):
    '''Generates and outputs COMBINE OMEX file.
    
    :param core_model: core model database of model
    :param settings: settings database of model
    :param Plot_Variable: Variables to be plotted in list format
    :param outputpath:  filepath to output OMEX file in OS path format
    :param KISAO_algorithm:  Kinetic Simulation Algorithm Ontology number in string format 
    :return combine_filename : OMEX file name in string format
    :return total_scenarios : Total number of scenarios in string format
    
    '''
    
    #--- Extract data from settings and clean up units ---
    # number_init         = len(settings['init'])
    number_init         = settings['init'].index
    number_parameters   = len(settings['parameters'])
    addparam            = settings['parameters']
    unit_model          = sbmlgen.unitlookup(settings)
    total_scenarios     = len(number_init) * number_parameters
    #print(total_scenarios)
    scenario_name       = core_model["system_type"]
    scenario_name       = scenario_name.replace(", ", "_")
    modelname_file      = []
    antimony_final      = []
    
    #--- Generates SBML string and adds to list
    #based on the number of combinations with init 
    #and parameter amts ---
    for j in number_init: #Cycle through number of init
        for k in range(number_parameters): #Cycle through number of parameters
            sbmlstr = sbmlgen.SBMLcreation(core_model, settings, unit_model, addparam, j, k) 
            antimony_final.append(te.sbmlToAntimony(sbmlstr))
    
    #--- Generates the scenario name for each SBML string generated ---
    scenario_placeholder = []
    for l in range(total_scenarios):
        for j in number_init: #Cycle through number of init
            for k in range(number_parameters):
              temp_scenario_name = '_' + str(j+1)+ '_' + str(k+1)
              scenario_placeholder.append(temp_scenario_name)
              
    #--- Names the models in antimony string according the scenario ---         
    inline_omex = '\n'
    for l in range(total_scenarios):
        scenario_name       = scenario_name + scenario_placeholder[l]
        modelname_file.append(scenario_name)
        antimony_final[l]   = antimony_final[l].replace('*doc0', modelname_file[l])
        scenario_name       = scenario_name.replace(scenario_placeholder[l], "")
        inline_omex         = inline_omex + str(antimony_final[l])
        #print("This is antimony_final", scenario_number,": \n", antimony_final[l])
    
    #--- Generates phrasedml file ---    
    phrasedml_final = gen_phrasedml(settings, modelname_file, Plot_Variable, KISAO_algorithm)    
    
    #--- Generates and outputs OMEX file---
    inline_omex         = inline_omex + phrasedml_final
    combine_filename    = 'COMBINE_' + scenario_name + '.omex'
    
    archive = os.path.join(outputpath, combine_filename)
    print('COMBINE Archive output: ', archive)
    te.exportInlineOmex(inline_omex, archive)
    
    return combine_filename, total_scenarios


#--- phrasedml generator and its functions ---    
def gen_phrasedml(settings, modelname_file, Plot_Variable, KISAO_algorithm):
    '''
    Generates the phrasedml task statement.
    
    :param settings: settings database of model
    :param modelname_file: Scenario Model in list format
    :param Plot_Variable: Variables to be plotted in list format
    :param KISAO_algorithm:  Kinetic Simulation Algorithm Ontology number in string format 
    :return: phrasedml in string format
    
    '''
    
    model_statement = ""    
    tspan = settings['tspan']
    
    #--- Generates the tasks statements based on models and sims (number of models * number of sims) --- 
    task_statement = ""
    task_count = 1
    for model_name in range(len(modelname_file)):
        model_statement     = model_statement + 'model' + str(model_name+1) + ' = model "' +  modelname_file[model_name] + '"\n'
        for tasksim_count in range(len(tspan)):
            task_statement  = task_statement + 'task' + str(task_count)
            task_statement  = task_statement + ' = run sim' + str(tasksim_count+1) + ' on '
            task_count      += 1
            task_statement  = task_statement + 'model' + str(model_name+1) + '\n'
    
    print("Task_statement = \n", task_statement)    
    
    
    print('this is tspan length = ', len(tspan))
    
    #--- Generates simulation statements with tspan (depended on number of tspans declared) ---
    sim_statement = gen_simstatements(tspan, KISAO_algorithm)
    
    #--- Generates the plot variables for each figure ---  
    variabletask_list, task_total   = gen_plotvariables(modelname_file, tspan, Plot_Variable)
    
    #--- Groups variables pertaining to the same plot figure together ---                             
    variable_statement, Total_Fig   = clean_groupvariables(Plot_Variable, tspan, task_total, variabletask_list) 
      
    #--- Generates the 'plot figure' statements based on tasks and sims --- 
    Plot_statement = gen_figurestatment(Total_Fig, variable_statement, Plot_Variable)
    
    phrasedml_final = model_statement + sim_statement + task_statement + Plot_statement

    print("This is phrasedml_final: \n" , phrasedml_final)
    return phrasedml_final
 
def gen_simstatements(tspan, KISAO_algorithm):
    '''
    Generates simulation statements with tspan (depended on number of tspans declared)    
    
    :param tspan: tspan of model in list of arrays
    :param KISAO_algorithm:  Kinetic Simulation Algorithm Ontology number in string format 
    :return: simulations statements with KISAO algorithm in string format
    
    '''
    sim_statement = "" 
    for sim_count in range(len(tspan)): 
        tspan_start         = float(tspan[sim_count][0])
        tspan_end           = float(tspan[sim_count][-1])
        tspan_interval      = len(tspan[sim_count])
        sim_statement       = (sim_statement + 'sim' + str(sim_count+1) + ' = simulate uniform(' 
                               + str(tspan_start) + ", "+ str(tspan_end) + ", " + str(tspan_interval) + ')\n')
        if KISAO_algorithm != '0':
            sim_statement   = sim_statement + 'sim' + str(sim_count+1) + '.algorithm = ' + KISAO_algorithm + '\n'
    return sim_statement


def gen_plotvariables(modelname_file, tspan, Plot_Variable):
    '''
    Generates the plot variables for each figure  
    
    :param modelname_file: the name of model as per scenario in list format
    :param tspan: tspan of model in list of arrays
    :param Plot_Variable:  Variables to be plotted in list format
    :return variabletask_list: variables in list format 
    :return task_total: total number of tasks (number of sims * number of scenario models) in int format
    
    '''
    task_total = len(modelname_file) * len(tspan) 
    variabletask_list = []
    for taskvariable in range(task_total):
        for variable_count in range(len(Plot_Variable)):
            variabletask_list.append('task' + str(taskvariable+1) + '.' + Plot_Variable[variable_count])
    #print(variabletask_list, len(Plot_Variable))
    
    return variabletask_list, task_total



def clean_groupvariables(Plot_Variable, tspan, task_total, variabletask_list):
    '''
    Groups variables pertaining to the same plot figure together 
    
    :param Plot_Variable:  Variables to be plotted in list format
    :param tspan: tspan of model in list of arrays
    :param task_total: total number of tasks in int format
    :param variabletask_list: variables in list format
    :return variable_statement: variables for each plot in list format 
    :return Total_Fig: total number of plots (number of sims * number of variables to plot) in int format
    
    '''
    Total_Fig           = len(Plot_Variable) * len(tspan)
    variable_statement  = ['']*Total_Fig  
    var_stepjump        = len(Plot_Variable) * len(tspan)   
    fig_variable        = 0                                                
    for plot_count in range(Total_Fig):
        while fig_variable < (task_total*len(Plot_Variable)): 
            if (fig_variable+var_stepjump) < (task_total*len(Plot_Variable)):
                variable_statement[plot_count] = variable_statement[plot_count] + variabletask_list[fig_variable] + ', ' 
            else:
                variable_statement[plot_count] = variable_statement[plot_count] + variabletask_list[fig_variable]
            fig_variable    = fig_variable + var_stepjump
        fig_variable        = plot_count + 1
      
    return variable_statement, Total_Fig
  
    
  
def gen_figurestatment(Total_Fig, variable_statement, Plot_Variable):
    '''
    Generates the 'plot figure' statements based on tasks and sims
    
    :param variable_statement: variables for each plot in list format 
    :param Total_Fig: total number of plots in int format
    :return Plot_statement: Plot figure statements in str format
    '''
    Figure_Statement    = []
    Plot_statement      = "" 
    time_count          = 1
    for plot_count in range(Total_Fig):#Change to meet fig count
        Figure_Statement.append('plot "Figure ' + str(plot_count+1) + '" task' + str(time_count) +'.time vs ')
        Plot_statement = Plot_statement + Figure_Statement[plot_count] + variable_statement[plot_count] + '\n'
        if (plot_count+1)%len(Plot_Variable) == 0:
            time_count +=1
    return Plot_statement

def KISAOchecker(KISAO_algorithm):
    '''
    Checks if KISAO_algorithm is under tellurium's accepted KISAO list'
    
    :param KISAO_algorithm:  Kinetic Simulation Algorithm Ontology number in string format
    :return: True if in tellurium list, false otherwise
    '''
    
    accepted_KISAO_list = ['kisao.0000019', 'kisao.0000433', 'kisao.0000407',
                           'kisao.0000099', 'kisao.0000035', 'kisao.0000071',
                           'kisao.0000288', 'kisao.0000280', 'kisao.0000032',
                           'kisao.0000064', 'kisao.0000435', 'kisao.0000321',
                           'kisao.0000087', 'kisao.0000086', 'kisao.0000434',
                           'kisao.0000088', 'kisao.0000241', 'kisao.0000029',
                           'kisao.0000319', 'kisao.0000274', 'kisao.0000333',
                           'kisao.0000329', 'kisao.0000323', 'kisao.0000331',
                           'kisao.0000027', 'kisao.0000082', 'kisao.0000324',
                           'kisao.0000350', 'kisao.0000330', 'kisao.0000028',
                           'kisao.0000038', 'kisao.0000039', 'kisao.0000048',
                           'kisao.0000074', 'kisao.0000081', 'kisao.0000045',
                           'kisao.0000351', 'kisao.0000084', 'kisao.0000040',
                           'kisao.0000046', 'kisao.0000003', 'kisao.0000051',
                           'kisao.0000335', 'kisao.0000336', 'kisao.0000095',
                           'kisao.0000022', 'kisao.0000076', 'kisao.0000015',
                           'kisao.0000075', 'kisao.0000278', 'kisao.0000099',
                           'kisao.0000274', 'kisao.0000282', 'kisao.0000283',
                           'kisao.0000355', 'kisao.0000356', 'kisao.0000407',
                           'kisao.0000408', 'kisao.0000409', 'kisao.0000410',
                           'kisao.0000411', 'kisao.0000412', 'kisao.0000413',
                           'kisao.0000432', 'kisao.0000437', '0']
    
    if KISAO_algorithm not in accepted_KISAO_list:
        raise Exception('KISAO Algorithm not accepted in Tellurium')
    
    return

        
def plotvariablechecker(Plot_Variable, core_model):
    '''
    :meta private:
    '''
    if len(Plot_Variable) == 0:
        raise Exception('No Plot Variable was declared.')
    for variable in Plot_Variable:
        if variable not in core_model['states']:
            raise NameError('Plot Variable declared does not exist in model')  
    return
    
    


    

    
    


import numpy  as np
import pandas as pd

def read_data(states_and_filenames, n_models=1):
    data_mu   = {}
    data_sd   = {}
    state_sd  = {}
    tspan     = []
    init      = {}
    
    for state, filename in states_and_filenames.items():
        df         = pd.read_csv(filename)
        scenarios  = []
        
        for column in df.columns:
            if 'std' in column:
                continue
            elif 'Time' in column:
                scenarios = [column] + scenarios
                tspan.append(df[column].values)
            else:
                scenarios.append(column)
              
        #Set up data_mu, data_sd, init, tspan
        for model_num in range(1, n_models+1):
            data_mu.setdefault(model_num, {})[state] = {}
            data_sd.setdefault(model_num, {})[state] = {}
            init.setdefault(model_num, {})
            
            for i, scenario in enumerate(scenarios):
                
                #Add experimental data and time to data_mu
                data_mu[model_num][state][i] = df[scenario].values    
                
                if 'Time' in scenario:
                    #Add time to data_sd
                    data_sd[model_num][state][i] = df[scenario].values
                
                else:
                    #Add sd of experimental data to data_sd
                    data_sd[model_num][state][i] = df[scenario + 'std'].values         
                    
                    #Set up initial values for integration
                    #Take first value of experimental data and add to init
                    init_val           = data_mu[model_num][state][i][0]              
                    init[model_num][i] = {state:init_val}
        
            #Add scenarios for reference    
            data_mu[model_num][state][-1] = scenarios
            data_sd[model_num][state][-1] = scenarios
        
        #Set up state_sd
        df_sd           = df[[scenario + 'std' for scenario in scenarios if 'Time' not in scenario]]
        state_sd[state] = df_sd.mean().mean()
    
    #Reformat tspan into a 1-D array and sort uniquely.
    tspan = np.unique(np.concatenate(tspan))
    return data_mu, data_sd, init, state_sd, tspan